# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import mock
import pytest

from kafka_utils.kafka_check import status_code
from kafka_utils.kafka_check.commands.min_isr import MinIsrCmd


class TestMinIsrCommand(object):
    REGULAR_TOPIC_NAME0 = 'topic_0'
    REGULAR_TOPIC_NAME1 = 'topic_1'
    TOPICS_STATE = {
        REGULAR_TOPIC_NAME0: {
            "version": 1,
            "partitions": {
                "0": {
                    "controller_epoch": 64,
                    "leader_epoch": 54,
                    "replicas": [
                        170396635,
                        170398981
                    ],
                    "isr": [
                        170398981
                    ],
                    "version": 1,
                    "leader": 170396635
                }
            }
        },
        REGULAR_TOPIC_NAME1: {
            "version": 1,
            "partitions": {
                "0": {
                    "controller_epoch": 64,
                    "leader_epoch": 189,
                    "replicas": [
                        170396635,
                        170398981
                    ],
                    "isr": [
                        170396635,
                        170398981
                    ],
                    "version": 1,
                    "leader": 170396635
                }
            }
        },
    }

    @pytest.fixture
    def cmd(self):
        return MinIsrCmd()

    def _prepare_topics(self, cmd, topics, min_isr, default_min_isr=None):
        attrs = {'get_topics.return_value': topics}
        cmd.zk = mock.MagicMock(**attrs)
        cmd.get_min_isr = mock.MagicMock(return_value=min_isr)
        cmd.args = mock.MagicMock(**{'default_min_isr': default_min_isr})

    def test_get_min_isr_empty(self, cmd):
        TOPIC_CONFIG_WITHOUT_MIN_ISR = {'version': 1, 'config': {'retention.ms': '86400000'}}
        attrs = {'get_topic_config.return_value': TOPIC_CONFIG_WITHOUT_MIN_ISR}
        cmd.zk = mock.MagicMock(**attrs)

        min_isr = cmd.get_min_isr(self.REGULAR_TOPIC_NAME0)

        assert cmd.zk.method_calls == [mock.call.get_topic_config(self.REGULAR_TOPIC_NAME0)]
        assert min_isr is None

    def test_get_min_isr(self, cmd):
        TOPIC_CONFIG_WITH_MIN_ISR = {
            'version': 1,
            'config': {'retention.ms': '86400000'},
            'min.insync.replicas': '3',
        }
        attrs = {'get_topic_config.return_value': TOPIC_CONFIG_WITH_MIN_ISR}
        cmd.zk = mock.MagicMock(**attrs)

        min_isr = cmd.get_min_isr(self.REGULAR_TOPIC_NAME0)

        assert cmd.zk.method_calls == [mock.call.get_topic_config(self.REGULAR_TOPIC_NAME0)]
        assert min_isr == 3

    def test_run_command_empty(self, cmd):
        self._prepare_topics(cmd, topics={}, min_isr=0)
        code, msg = cmd.run_command()

        cmd.get_min_isr.assert_not_called()
        assert code == status_code.OK
        assert msg == "All replicas in sync."

    def test_run_command_all_ok(self, cmd):
        self._prepare_topics(cmd, topics=self.TOPICS_STATE, min_isr=1)
        code, msg = cmd.run_command()

        assert code == status_code.OK
        assert msg == "All replicas in sync."

    def test_run_command_all_ok_without_min_isr_in_zk(self, cmd):
        self._prepare_topics(cmd, topics=self.TOPICS_STATE, min_isr=None)
        code, msg = cmd.run_command()

        assert code == status_code.OK
        assert msg == "All replicas in sync."

    def test_run_command_with_default_min_isr(self, cmd):
        self._prepare_topics(cmd, topics=self.TOPICS_STATE, min_isr=None, default_min_isr=1)
        code, msg = cmd.run_command()

        assert code == status_code.OK
        assert msg == "All replicas in sync."

    def test_run_command_with_fail_with_default_min_isr(self, cmd):
        self._prepare_topics(cmd, topics=self.TOPICS_STATE, min_isr=None, default_min_isr=2)
        code, msg = cmd.run_command()

        assert code == status_code.CRITICAL
        assert msg == "1 partition(s) have the number of replicas in sync that is lower than the specified min ISR."

    def test_run_command_all_fail(self, cmd):
        self._prepare_topics(cmd, topics=self.TOPICS_STATE, min_isr=3)
        code, msg = cmd.run_command()

        assert code == status_code.CRITICAL
        assert msg == "2 partition(s) have the number of replicas in sync that is lower than the specified min ISR."



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

from kafka_utils.kafka_cluster_manager.cluster_info.partition import Partition
from kafka_utils.kafka_cluster_manager.cluster_info.topic import Topic


@pytest.fixture
def create_partition():
    """Fixture to create a partition and attach it to a topic"""
    topics = {}

    def _add_partition(topic_id, partition_id, replication_factor=1):
        if topic_id not in topics:
            topics[topic_id] = Topic(topic_id, replication_factor)
        topic = topics[topic_id]
        partition = Partition(topic, partition_id)
        topic.add_partition(partition)
        return partition

    return _add_partition



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import contextlib

import mock
import pytest
from kazoo.exceptions import ZookeeperError

from kafka_utils.kafka_consumer_manager. \
    commands.rename_group import RenameGroup


@mock.patch(
    'kafka_utils.kafka_consumer_manager.'
    'commands.rename_group.KafkaToolClient',
    autospec=True,
)
class TestRenameGroup(object):

    @contextlib.contextmanager
    def mock_kafka_info(self, topics_partitions):
        with mock.patch.object(
            RenameGroup,
            "preprocess_args",
            spec=RenameGroup.preprocess_args,
            return_value=topics_partitions,
        ) as mock_process_args, mock.patch(
            "kafka_utils.kafka_consumer_manager.util.prompt_user_input",
            autospec=True,
        ) as mock_user_confirm, mock.patch(
            "kafka_utils.kafka_consumer_manager."
            "commands.rename_group.ZK",
            autospec=True,
        ) as mock_ZK:
            mock_ZK.return_value.__enter__.return_value = mock_ZK
            yield mock_process_args, mock_user_confirm, mock_ZK

    def test_run(self, mock_client):
        topics_partitions = {
            "topic1": [0, 1, 2],
            "topic2": [0, 1]
        }
        with self.mock_kafka_info(
            topics_partitions
        ) as (mock_process_args, mock_user_confirm, mock_ZK):
            obj = mock_ZK.return_value.__enter__.return_value
            obj.get_children.return_value = [
                'some_topic', 'another_topic'
            ]
            obj.get.return_value = (0, 0)
            cluster_config = mock.Mock(zookeeper='some_ip')
            args = mock.Mock(old_groupid='old_group', new_groupid='new_group')
            zk_old_group_get_calls = [
                mock.call("/consumers/old_group/offsets/topic1/0"),
                mock.call("/consumers/old_group/offsets/topic1/1"),
                mock.call("/consumers/old_group/offsets/topic1/2"),
                mock.call("/consumers/old_group/offsets/topic2/0"),
                mock.call("/consumers/old_group/offsets/topic2/1"),
            ]
            zk_old_group_delete_calls = [
                mock.call("/consumers/old_group", recursive=True),
            ]
            zk_new_group_calls = [
                mock.call(
                    "/consumers/new_group/offsets/topic1/0",
                    value=bytes(0),
                    makepath=True
                ),
                mock.call(
                    "/consumers/new_group/offsets/topic1/1",
                    value=bytes(0),
                    makepath=True
                ),
                mock.call(
                    "/consumers/new_group/offsets/topic1/2",
                    value=bytes(0),
                    makepath=True
                ),
                mock.call(
                    "/consumers/new_group/offsets/topic2/0",
                    value=bytes(0),
                    makepath=True
                ),
                mock.call(
                    "/consumers/new_group/offsets/topic2/1",
                    value=bytes(0),
                    makepath=True
                ),
            ]

            RenameGroup.run(args, cluster_config)

            assert mock_user_confirm.call_count == 1
            sorted(obj.get.call_args_list) == sorted(zk_old_group_get_calls)
            sorted(obj.delete.call_args_list) == sorted(zk_old_group_delete_calls)
            sorted(obj.create.call_args_list) == sorted(zk_new_group_calls)

    def test_run_same_groupids(self, mock_client):
        topics_partitions = {}
        with self.mock_kafka_info(
            topics_partitions
        ) as (mock_process_args, mock_user_confirm, mock_ZK):
            with pytest.raises(SystemExit) as ex:
                cluster_config = mock.Mock(zookeeper='some_ip')
                args = mock.Mock(
                    old_groupid='my_group',
                    new_groupid='my_group',
                )

                RenameGroup.run(args, cluster_config)

                ex.value.code == 1

    def test_run_topic_already_subscribed_to_error(self, mock_client):
        topics_partitions = {
            "topic1": [0, 1, 2],
            "topic2": [0, 1]
        }
        with self.mock_kafka_info(
            topics_partitions
        ) as (mock_process_args, mock_user_confirm, mock_ZK):
            with pytest.raises(SystemExit) as ex:
                obj = mock_ZK.return_value.__enter__.return_value
                cluster_config = mock.Mock(zookeeper='some_ip')
                args = mock.Mock(
                    old_groupid='old_group',
                    new_groupid='new_group',
                )
                obj.get_children.return_value = ['topic1']
                obj.get.return_value = (0, 0)

                RenameGroup.run(args, cluster_config)

                assert ex.value.code == 1

    def test_run_create_zknode_error(self, mock_client):
        topics_partitions = {
            "topic1": [0, 1, 2],
            "topic2": [0, 1]
        }
        with self.mock_kafka_info(
            topics_partitions
        ) as (mock_process_args, mock_user_confirm, mock_ZK):
            obj = mock_ZK.return_value.__enter__.return_value
            obj.__exit__.return_value = False
            cluster_config = mock.Mock(zookeeper='some_ip')
            args = mock.Mock(old_groupid='old_group', new_groupid='new_group')
            obj.get_children.return_value = [
                'some_topic', 'another_topic'
            ]
            obj.get.return_value = (0, 0)
            obj.create.side_effect = ZookeeperError("Boom!")

            with pytest.raises(ZookeeperError):
                RenameGroup.run(args, cluster_config)
            assert mock_user_confirm.call_count == 1

    def test_run_delete_zknode_error(self, mock_client):
        topics_partitions = {
            "topic1": [0, 1, 2],
            "topic2": [0, 1]
        }
        with self.mock_kafka_info(
            topics_partitions
        ) as (mock_process_args, mock_user_confirm, mock_ZK):
            obj = mock_ZK.return_value.__enter__.return_value
            obj.__exit__.return_value = False
            cluster_config = mock.Mock(zookeeper='some_ip')
            args = mock.Mock(old_groupid='old_group', new_groupid='new_group')
            obj.get_children.return_value = []
            obj.get.return_value = (0, 0)
            obj.delete.side_effect = ZookeeperError("Boom!")

            with pytest.raises(ZookeeperError):
                RenameGroup.run(args, cluster_config)
            assert mock_user_confirm.call_count == 0



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging

from kafka_utils.util.validation import assignment_to_plan
_log = logging.getLogger('kafka-cluster-manager')


def display_cluster_topology(cluster_topology):
    print(assignment_to_plan(cluster_topology.assignment))


def display_same_replica_count_rg(
    duplicate_replica_count_per_rg,
    imbalance,
):
    """Display same topic/partition count over brokers."""
    print("=" * 40)
    print("Replication-group Extra-Same-replica-count")
    print("=" * 40)
    for rg_id, replica_count in duplicate_replica_count_per_rg.iteritems():
        count = int(replica_count)
        print("{b:^7s} {cnt:^10d}".format(b=rg_id, cnt=int(count)))
    print("=" * 40)
    print(
        '\nTotal replication-group imbalance: {imbalance}\n\n'
        .format(imbalance=imbalance),
    )


def display_same_topic_partition_count_broker(
    same_topic_partition_count,
    imbalance,
):
    """Display same topic/partition count over brokers."""
    print("=" * 40)
    print("Broker   Extra-Topic-Partition Count")
    print("=" * 40)
    for broker_id, partition_count in \
            same_topic_partition_count.iteritems():
        print(
            "{b:^7d} {partition_count:^18d}"
            .format(b=broker_id, partition_count=partition_count),
        )
    print("=" * 40)
    print(
        '\nTotal same topic-partition imbalance count: {imbalance}\n'
        .format(imbalance=imbalance),
    )


def display_partition_count_per_broker(
    partition_count,
    stdev_imbalance,
    imbalance,
):
    """Display partition count over brokers."""
    print("=" * 25)
    print("Broker   Partition Count")
    print("=" * 25)
    for broker_id, count in partition_count.iteritems():
        print(
            "{b:^7d} {cnt:^18d}".format(
                b=broker_id,
                cnt=count,
            ),
        )
    print("=" * 25)

    print('\n\nPartition-count imbalance:')
    print('Standard-deviation: {stdev}'.format(stdev=stdev_imbalance))
    print(
        'Net-imbalance: Total extra partitions over brokers: {imbalance}'
        .format(imbalance=imbalance),
    )


def display_leader_count_per_broker(leader_count, stdev_imbalance, imbalance):
    """Display partition and leader count for each broker."""
    print("=" * 33)
    print("Broker   Preferred Leader Count")
    print("=" * 33)
    for broker_id, count in leader_count.iteritems():
        print(
            "{b:^7d} {cnt:^20d}".format(
                b=broker_id,
                cnt=count,
            ),
        )
    print("=" * 33)
    print('\n\nLeader-count imbalance:')
    print('Standard-deviation: {stdev}'.format(stdev=stdev_imbalance))
    print(
        'Net-imbalance: Total extra brokers as leaders: {imbalance}'
        .format(imbalance=imbalance),
    )


def display_assignment_changes(plan_details, to_log=True):
    """Display current and proposed changes in
    topic-partition to replica layout over brokers.
    """
    curr_plan_list, new_plan_list, total_changes = plan_details
    action_cnt = '\n[INFO] Total actions required {0}'.format(total_changes)
    _log_or_display(to_log, action_cnt)
    action_cnt = (
        '[INFO] Total actions that will be executed {0}'
        .format(len(new_plan_list))
    )
    _log_or_display(to_log, action_cnt)
    changes = ('[INFO] Proposed Changes in current cluster-layout:\n')
    _log_or_display(to_log, changes)

    tp_str = 'Topic - Partition'
    curr_repl_str = 'Previous-Assignment'
    new_rep_str = 'Proposed-Assignment'
    tp_list = [tp_repl[0] for tp_repl in curr_plan_list]

    # Display heading
    msg = '=' * 80
    _log_or_display(to_log, msg)
    row = (
        '{tp:^30s}: {curr_rep_str:^20s} ==> {new_rep_str:^20s}' .format(
            tp=tp_str,
            curr_rep_str=curr_repl_str,
            new_rep_str=new_rep_str,
        )
    )
    _log_or_display(to_log, row)
    msg = '=' * 80
    _log_or_display(to_log, msg)

    # Display each topic-partition list with changes
    tp_list_sorted = sorted(tp_list, key=lambda tp: (tp[0], tp[1]))
    for tp in tp_list_sorted:
        curr_repl = [
            tp_repl[1] for tp_repl in curr_plan_list if tp_repl[0] == tp
        ][0]
        proposed_repl = [
            tp_repl[1] for tp_repl in new_plan_list if tp_repl[0] == tp
        ][0]
        tp_str = '{topic} - {partition:<2d}'.format(topic=tp[0], partition=tp[1])
        row = (
            '{tp:<30s}: {curr_repl:<20s} ==> {proposed_repl:<20s}'.format(
                tp=tp_str,
                curr_repl=curr_repl,
                proposed_repl=proposed_repl,
            )
        )
        _log_or_display(to_log, row)


def _log_or_display(to_log, msg):
    """Log or display the information."""
    if to_log:
        _log.info(msg)
    else:
        print(msg)



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import struct

from kafka.protocol import KafkaProtocol
from kafka.util import group_by_topic_and_partition
from kafka.util import write_short_string


class KafkaToolProtocol(KafkaProtocol):

    @classmethod
    def encode_offset_commit_request_kafka(cls, client_id, correlation_id,
                                           group, payloads):
        """
        Encode some OffsetCommitRequest structs
        Arguments:
            client_id: string
            correlation_id: int
            group: string, the consumer group you are committing offsets for
            payloads: list of OffsetCommitRequest
        """
        grouped_payloads = group_by_topic_and_partition(payloads)

        message = []
        message.append(cls._encode_message_header(
            client_id, correlation_id,
            KafkaProtocol.OFFSET_COMMIT_KEY,
            version=2))
        message.append(write_short_string(group))
        message.append(struct.pack('>i', -1))   # ConsumerGroupGenerationId
        message.append(write_short_string(''))  # ConsumerId
        message.append(struct.pack('>q', -1))   # Retention time
        message.append(struct.pack('>i', len(grouped_payloads)))

        for topic, topic_payloads in grouped_payloads.items():
            message.append(write_short_string(topic))
            message.append(struct.pack('>i', len(topic_payloads)))

            for partition, payload in topic_payloads.items():
                message.append(struct.pack('>iq', partition, payload.offset))
                message.append(write_short_string(payload.metadata))

        msg = b''.join(message)
        return struct.pack('>i%ds' % len(msg), len(msg), msg)



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import sys
from collections import defaultdict

from kazoo.exceptions import NodeExistsError


def preprocess_topics(source_groupid, source_topics, dest_groupid, topics_dest_group):
    """Pre-process the topics in source and destination group for duplicates."""
    # Is the new consumer already subscribed to any of these topics?
    common_topics = [topic for topic in topics_dest_group if topic in source_topics]
    if common_topics:
        print(
            "Error: Consumer Group ID: {groupid} is already "
            "subscribed to following topics: {topic}.\nPlease delete this "
            "topics from new group before re-running the "
            "command.".format(
                groupid=dest_groupid,
                topic=', '.join(common_topics),
            ),
            file=sys.stderr,
        )
        sys.exit(1)
    # Let's confirm what the user intends to do.
    if topics_dest_group:
        in_str = (
            "New Consumer Group: {dest_groupid} already "
            "exists.\nTopics subscribed to by the consumer groups are listed "
            "below:\n{source_groupid}: {source_group_topics}\n"
            "{dest_groupid}: {dest_group_topics}\nDo you intend to copy into"
            "existing consumer destination-group? (y/n)".format(
                source_groupid=source_groupid,
                source_group_topics=source_topics,
                dest_groupid=dest_groupid,
                dest_group_topics=topics_dest_group,
            )
        )
        prompt_user_input(in_str)


def create_offsets(zk, consumer_group, offsets):
    """Create path with offset value for each topic-partition of given consumer
    group.

    :param zk: Zookeeper client
    :param consumer_group: Consumer group id for given offsets
    :type consumer_group: int
    :param offsets: Offsets of all topic-partitions
    :type offsets: dict(topic, dict(partition, offset))
    """
    # Create new offsets
    for topic, partition_offsets in offsets.iteritems():
        for partition, offset in partition_offsets.iteritems():
            new_path = "/consumers/{groupid}/offsets/{topic}/{partition}".format(
                groupid=consumer_group,
                topic=topic,
                partition=partition,
            )
            try:
                zk.create(new_path, value=bytes(offset), makepath=True)
            except NodeExistsError:
                print(
                    "Error: Path {path} already exists. Please re-run the "
                    "command.".format(path=new_path),
                    file=sys.stderr,
                )
                raise


def fetch_offsets(zk, consumer_group, topics):
    """Fetch offsets for given topics of given consumer group.

    :param zk: Zookeeper client
    :param consumer_group: Consumer group id for given offsets
    :type consumer_group: int
    :rtype: dict(topic, dict(partition, offset))
    """
    source_offsets = defaultdict(dict)
    for topic, partitions in topics.iteritems():
        for partition in partitions:
            offset, _ = zk.get(
                "/consumers/{groupid}/offsets/{topic}/{partition}".format(
                    groupid=consumer_group,
                    topic=topic,
                    partition=partition,
                )
            )
            source_offsets[topic][partition] = offset
    return source_offsets


def prompt_user_input(in_str):
    while(True):
        answer = raw_input(in_str + ' ')
        if answer == "n" or answer == "no":
            sys.exit(0)
        if answer == "y" or answer == "yes":
            return



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
from collections import defaultdict

import mock
import pytest

from kafka_utils.kafka_consumer_manager. \
    commands.offset_set import OffsetSet
from kafka_utils.util.error import OffsetCommitError


class TestOffsetSet(object):

    def test_topics_dict(self):
        offset_update_tuple = "topic1.23=1000"
        expected_topics_dict = {
            "topic1": {23: 1000},
        }
        OffsetSet.topics_dict(offset_update_tuple)
        assert OffsetSet.new_offsets_dict == expected_topics_dict

    def test_topics_dict_topic_with_period(self):
        OffsetSet.new_offsets_dict = defaultdict(dict)
        offset_update_tuple = "scribe.sfo2.ranger.12=200"
        expected_topics_dict = {
            "scribe.sfo2.ranger": {12: 200},
        }
        OffsetSet.topics_dict(offset_update_tuple)
        assert OffsetSet.new_offsets_dict == expected_topics_dict

    def test_topics_dict_invalid_input(self):
        offset_update_tuple = "topic1.23.1000"
        with mock.patch.object(sys, "exit", autospec=True) as mock_exit:
            OffsetSet.topics_dict(offset_update_tuple)
            mock_exit.assert_called_once_with(1)

        offset_update_tuple = "topic1.garbage=garbage"
        with mock.patch.object(sys, "exit", autospec=True) as mock_exit:
            OffsetSet.topics_dict(offset_update_tuple)
            mock_exit.assert_called_once_with(1)

    @mock.patch(
        'kafka_utils.kafka_consumer_manager.'
        'commands.offset_set.KafkaToolClient',
        autospec=True,
    )
    def test_run(self, mock_client):
        OffsetSet.new_offsets_dict = {
            "topic1": {
                0: 1000,
                1: 2000,
                2: 3000,
            },
            "topic2": {
                0: 100,
                1: 200,
            },
        }

        with mock.patch.object(
            OffsetSet,
            'get_topics_from_consumer_group_id',
            spec=OffsetSet.get_topics_from_consumer_group_id,
        ), mock.patch(
            "kafka_utils.kafka_consumer_manager."
            "commands.offset_set.set_consumer_offsets",
            return_value=[],
            autospec=True
        ) as mock_set_offsets:
            args = mock.Mock(
                groupid="some_group",
                topic=None,
                partitions=None
            )
            cluster_config = mock.Mock()
            OffsetSet.run(args, cluster_config)

            mock_client.return_value.load_metadata_for_topics. \
                assert_called_once_with()
            mock_client.return_value.close.assert_called_once_with()
            ordered_args, _ = mock_set_offsets.call_args
            assert ordered_args[1] == args.groupid
            assert ordered_args[2] == OffsetSet.new_offsets_dict

    @mock.patch(
        'kafka_utils.kafka_consumer_manager.'
        'commands.offset_set.KafkaToolClient',
        autospec=True,
    )
    def test_run_error_committing_offsets(self, mock_client):
        OffsetSet.new_offsets_dict = {
            "topic1": {
                0: 1000,
                1: 2000,
            },
            "topic2": {
                0: 100,
            },
        }

        with mock.patch.object(
            OffsetSet,
            'get_topics_from_consumer_group_id',
            spec=OffsetSet.get_topics_from_consumer_group_id,
        ), mock.patch(
            "kafka_utils.kafka_consumer_manager."
            "commands.offset_set.set_consumer_offsets",
            return_value=[
                OffsetCommitError("topic1", 1, "my_error 1"),
                OffsetCommitError("topic2", 0, "my_error 2"),
            ],
            autospec=True
        ) as mock_set_offsets, mock.patch.object(
            sys,
            "exit",
            autospec=True,
        ) as mock_exit:
            args = mock.Mock(
                groupid="some_group",
                topic=None,
                partitions=None
            )
            cluster_config = mock.Mock()
            OffsetSet.run(args, cluster_config)

            mock_client.return_value.load_metadata_for_topics. \
                assert_called_once_with()
            mock_client.return_value.close.assert_called_once_with()
            ordered_args, _ = mock_set_offsets.call_args
            assert ordered_args[1] == args.groupid
            assert ordered_args[2] == OffsetSet.new_offsets_dict
            mock_exit.assert_called_with(1)

    @mock.patch(
        'kafka_utils.kafka_consumer_manager.'
        'commands.offset_set.KafkaToolClient',
        autospec=True,
    )
    def test_run_bad_topics_dict(self, mock_client):
        OffsetSet.new_offsets_dict = {
            "topic1": 23,
            "topic2": 32,
        }
        with mock.patch.object(
            OffsetSet,
            'get_topics_from_consumer_group_id',
            spec=OffsetSet.get_topics_from_consumer_group_id,
        ):
            args = mock.Mock(
                groupid="some_group",
                topic=None,
                partitions=None
            )
            cluster_config = mock.Mock()
            with pytest.raises(TypeError):
                OffsetSet.run(args, cluster_config)



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from kafka_utils.kafka_cluster_manager.cluster_info.broker import Broker
from kafka_utils.kafka_cluster_manager.cluster_info.partition import Partition


def create_broker(broker_id, partitions):
    b = Broker(broker_id, partitions=set(partitions))
    for p in partitions:
        p.add_replica(b)
    return b


def create_and_attach_partition(topic, partition_id):
    partition = Partition(topic, partition_id)
    topic.add_partition(partition)
    return partition



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import mock
import pytest

from kafka_utils.kafka_consumer_manager. \
    commands.offset_get import OffsetGet


class TestOffsetGet(object):

    @pytest.yield_fixture
    def client(self):
        with mock.patch(
                'kafka_utils.kafka_consumer_manager.'
                'commands.offset_get.KafkaToolClient',
                autospec=True,
        ) as mock_client:
            yield mock_client

    def test_get_offsets(self, client):
        consumer_group = 'group1'
        topics = {'topic1': {0: 100}}

        with mock.patch(
            'kafka_utils.util.offsets._verify_topics_and_partitions',
            return_value=topics,
            autospec=True,
        ):
            OffsetGet.get_offsets(
                client,
                consumer_group,
                topics,
                'zookeeper',
            )

            assert client.load_metadata_for_topics.call_count == 1
            assert client.send_offset_fetch_request.call_count == 1
            assert client.send_offset_fetch_request_kafka.call_count == 0

    def test_get_offsets_kafka(self, client):
        consumer_group = 'group1'
        topics = {'topic1': {0: 100}}

        with mock.patch(
            'kafka_utils.util.offsets._verify_topics_and_partitions',
            return_value=topics,
            autospec=True,
        ):
            OffsetGet.get_offsets(
                client,
                consumer_group,
                topics,
                'kafka',
            )

            assert client.load_metadata_for_topics.call_count == 1
            assert client.send_offset_fetch_request.call_count == 0
            assert client.send_offset_fetch_request_kafka.call_count == 1



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import
from __future__ import print_function

from kafka_utils.kafka_check import status_code
from kafka_utils.kafka_check.commands.command import KafkaCheckCmd


class MinIsrCmd(KafkaCheckCmd):

    def build_subparser(self, subparsers):
        subparser = subparsers.add_parser(
            'min_isr',
            description='Check min isr number for each topic in the cluster.',
            help='This command will check actual number of insync replicas for each '
                 'topic-partition in the cluster with configuration for that topic '
                 'in Zookeeper or default min.isr param if it is specified and there '
                 'is no settings in Zookeeper for partition.',
        )
        subparser.add_argument(
            '--default_min_isr',
            type=int,
            default=1,
            help='Default min.isr value for cases without settings in Zookeeper '
            'for some topics. Default: %(default)s',
        )
        return subparser

    def get_min_isr(self, topic):
        """Return the min-isr for topic, or None if not specified"""
        ISR_CONF_NAME = 'min.insync.replicas'
        config = self.zk.get_topic_config(topic)
        if ISR_CONF_NAME in config:
            return int(config[ISR_CONF_NAME])
        else:
            return None

    def run_command(self):
        """Min_isr command, checks number of actual min-isr
        for each topic-partition with configuration for that topic."""
        topics = self.zk.get_topics()
        not_in_sync = 0

        for name, topic_data in topics.items():
            min_isr = self.get_min_isr(name) or self.args.default_min_isr
            if min_isr is None:
                continue
            for p_id, partition in topic_data['partitions'].items():
                cur_isr = len(partition['isr'])
                if cur_isr < min_isr:
                    print("isr={isr} is lower than min_isr={min_isr} for {topic}:{partition}".format(
                        isr=cur_isr, min_isr=min_isr, topic=name, partition=p_id,
                    ))
                    not_in_sync += 1
        if not_in_sync == 0:
            return status_code.OK, "All replicas in sync."
        else:
            msg = ("{0} partition(s) have the number of replicas in "
                   "sync that is lower than the specified min ISR.").format(not_in_sync)
            return status_code.CRITICAL, msg



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import mock
import pytest

from kafka_utils.kafka_cluster_manager.cluster_info \
    .cluster_topology import ClusterTopology
from kafka_utils.kafka_cluster_manager.cluster_info \
    .error import BrokerDecommissionError
from kafka_utils.kafka_cluster_manager.cluster_info \
    .error import InvalidBrokerIdError
from kafka_utils.kafka_cluster_manager.cluster_info \
    .error import InvalidPartitionError
from kafka_utils.kafka_cluster_manager.cluster_info \
    .error import RebalanceError
from kafka_utils.kafka_cluster_manager.cluster_info \
    .stats import calculate_partition_movement
from kafka_utils.kafka_cluster_manager.cluster_info \
    .stats import get_leader_imbalance_stats
from kafka_utils.kafka_cluster_manager.cluster_info \
    .stats import get_replication_group_imbalance_stats


class TestClusterTopology(object):
    # replication-group to broker map
    # rg1: 0, 1, 4; rg2: 2, 3; rg3: 5; rg4: 6;
    broker_rg = {
        '0': 'rg1', '1': 'rg1', '2': 'rg2', '3': 'rg2',
        '4': 'rg1', '5': 'rg3', '6': 'rg4',
    }
    topic_ids = ['T0', 'T1', 'T2', 'T3']
    brokers = {
        '0': {'host': 'host1'},
        '1': {'host': 'host2'},
        '2': {'host': 'host3'},
        '3': {'host': 'host4'},
        '4': {'host': 'host5'},
    }
    # Example assignment properties:
    # * Brokers:(0,1,2,3): rg-count = 2
    # case 1: replication-factor % rg-count == 0
    #    -- T0, T1:
    #   * 1a) T1: replication-factor > rg-count
    # case 2: replication-factor % rg-count != 0
    #   -- T2, T3
    #   * 2a): replication-factor > rg-count: T1
    #   * 2b): replication-factor < rg-count: T2
    # rg-imbalance-status per partition:
    #
    # rg-imbalanced-partitions: T0-1, T1-1, T3-1
    # rg-balanced-partitions:   T0-0, T1-0, T3-0, T2-0
    _initial_assignment = dict(
        [
            ((u'T0', 0), ['1', '2']),
            ((u'T0', 1), ['2', '3']),
            ((u'T1', 0), ['0', '1', '2', '3']),
            ((u'T1', 1), ['0', '1', '2', '4']),
            ((u'T2', 0), ['2']),
            ((u'T3', 0), ['0', '1', '2']),
            ((u'T3', 1), ['0', '1', '4']),
        ]
    )

    def get_replication_group_id(self, broker):
        try:
            return self.broker_rg[broker.id]
        except KeyError:  # inactive brokers
            return None

    def build_cluster_topology(self, assignment=None, brokers=None):
        """Create cluster topology from given assignment."""
        if not assignment:
            assignment = self._initial_assignment
        if not brokers:
            brokers = self.brokers
        return ClusterTopology(assignment, brokers, self.get_replication_group_id)

    def test_cluster_topology_inactive_brokers(self):
        assignment = {
            (u'T0', 0): ['0', '1'],
            (u'T0', 1): ['8', '9'],  # 8 and 9 are not in active brokers
        }
        brokers = {
            '0': {'host': 'host0'},
            '1': {'host': 'host1'},
        }

        def extract_group(broker):
            # group 0 for broker 0
            # group 1 for broker 1
            # None for inactive brokers
            if broker in brokers:
                return broker.id
            return None

        ct = ClusterTopology(assignment, brokers, extract_group)
        assert ct.brokers['8'].inactive
        assert ct.brokers['9'].inactive
        assert None in ct.rgs

    def test_broker_decommission(self):
        assignment = {
            (u'T0', 0): ['0', '2'],
            (u'T0', 1): ['0', '3'],
            (u'T1', 0): ['0', '5'],
        }
        ct = self.build_cluster_topology(assignment)
        partitions_count = len(ct.partitions)

        # should move all partitions from broker 0 to either 1 or 4 because they
        # are in the same replication group and empty.
        ct.decommission_brokers(['0'])

        # Here we just care that broker 0 is empty and partitions count didn't
        # change
        assert len(ct.partitions) == partitions_count
        assert ct.brokers['0'].empty()

    def test_broker_decommission_many(self):
        assignment = {
            (u'T0', 0): ['0', '2'],
            (u'T0', 1): ['0', '3'],
            (u'T1', 0): ['0', '5'],
            (u'T1', 1): ['1', '5'],
            (u'T1', 2): ['1', '5'],
            (u'T2', 0): ['0', '3'],
        }
        ct = self.build_cluster_topology(assignment)
        partitions_count = len(ct.partitions)

        ct.decommission_brokers(['0', '1', '3'])

        assert len(ct.partitions) == partitions_count
        assert ct.brokers['0'].empty()
        assert ct.brokers['1'].empty()
        assert ct.brokers['3'].empty()
        # All partitions from 0 and 1 should now be in 4
        assert len(ct.brokers['4'].partitions) == 6
        # All partitions from 3 should now be in 2
        assert len(ct.brokers['2'].partitions) == 3

    def test_broker_decommission_failover(self):
        assignment = {
            (u'T0', 0): ['0', '1', '2'],
            (u'T0', 1): ['0', '1', '2'],
            (u'T1', 0): ['0', '1'],
            (u'T1', 1): ['1', '5'],
            (u'T1', 2): ['1', '5'],
            (u'T2', 0): ['0', '3'],
        }
        ct = self.build_cluster_topology(assignment)
        partitions_count = len(ct.partitions)

        ct.decommission_brokers(['0'])

        # Partition T00, T01, and T10 should move to 5 and 6
        assert len(ct.partitions) == partitions_count
        assert ct.brokers['0'].empty()

    def test_broker_decommission_error(self):
        assignment = {
            (u'T1', 0): ['0', '1', '2', '3', '4'],
            (u'T1', 1): ['0', '1', '2', '4'],
            (u'T2', 1): ['0', '1', '2', '4'],
        }
        ct = self.build_cluster_topology(assignment)

        with pytest.raises(BrokerDecommissionError):
            ct.decommission_brokers(['0'])

    def test_rebalance_replication_groups(self):
        ct = self.build_cluster_topology()
        ct.rebalance_replication_groups()
        net_imbal, _ = get_replication_group_imbalance_stats(
            ct.rgs.values(),
            ct.partitions.values(),
        )

        # Verify that rg-group-balanced
        assert net_imbal == 0

        # Verify that new-assignment is valid
        self.assert_valid(
            ct.assignment,
            self._initial_assignment,
            ct.brokers.keys(),
        )

    def test_rebalance_replication_groups_balanced(self):
        # Replication-group is already balanced
        assignment = dict(
            [
                ((u'T0', 0), ['0', '2']),
                ((u'T0', 1), ['0', '3']),
                ((u'T2', 0), ['2']),
                ((u'T3', 0), ['0', '1', '2']),
            ]
        )
        ct = self.build_cluster_topology(assignment, self.srange(5))
        ct.rebalance_replication_groups()
        net_imbal, _ = get_replication_group_imbalance_stats(
            ct.rgs.values(),
            ct.partitions.values(),
        )

        # Verify that rg-group-balanced
        assert net_imbal == 0
        # Verify that new-assignment same as previous
        assert ct.assignment == assignment

    def test_rebalance_replication_groups_error(self):
        assignment = dict(
            [
                ((u'T0', 0), ['0', '2']),
                ((u'T0', 1), ['0', '3']),
                ((u'T2', 0), ['2']),
                ((u'T3', 0), ['0', '1', '9']),  # broker 9 is not active
            ]
        )
        ct = self.build_cluster_topology(assignment, self.srange(5))

        with pytest.raises(RebalanceError):
            ct.rebalance_replication_groups()

    def test_elect_source_replication_group(self):
        # Sample assignment with 3 replication groups
        # with replica-count as as per map :-
        # broker_rg = {0: 'rg1', 1: 'rg1', 2: 'rg2', 3: 'rg2', 4: 'rg1', 5: 'rg3', 6: 'rg4'}
        # rg-id:    (brokers), count
        # rg1:      (0, 2, 4) = 3
        # rg2:      (1, 3) = 2
        # rg3:      (5) = 1
        # rg4:      (6) = 1
        # rg1 and rg2 are over-replicated and rg3 being under-replicated
        # source-replication-group should be rg1 having the highest replicas
        p1 = ((u'T0', 0), ['0', '1', '2', '3', '4', '5', '6'])
        assignment = dict([p1])
        ct = self.build_cluster_topology(assignment, self.srange(7))
        # Case-1: rg's have only 1 unique max replica count
        # 'rg1' and 'rg2' are over-replicated groups
        over_replicated_rgs = [ct.rgs['rg1'], ct.rgs['rg2']]

        # Get source-replication group
        rg_source = ct._elect_source_replication_group(over_replicated_rgs, p1)

        # Since, 'rg1' has more replicas i.e. 3, it should be selected
        assert rg_source.id == 'rg1'

    def test_elect_dest_replication_group(self):
        # Sample assignment with 3 replication groups
        # with replica-count as as per map
        # broker_rg: {0: 'rg1', 1: 'rg1', 2: 'rg2', 3: 'rg2', 4: 'rg1', 5: 'rg3'}
        # rg-id: (brokers), count
        # rg1: (0, 2, 4) = 3
        # rg2: (1, 3) = 2
        # rg3: (5) = 1
        # rg1 and rg2 are over-replicated and rg3 being under-replicated
        # source-replication-group should be rg1 having the highest replicas
        p1_info = ((u'T0', 0), ['0', '1', '2', '3', '4', '5', '6'])
        assignment = dict([p1_info])
        ct = self.build_cluster_topology(assignment, self.srange(7))
        p1 = ct.partitions[p1_info[0]]
        # Case 1: rg_source = 'rg1', find destination-replica
        rg_source = ct.rgs['rg1']
        under_replicated_rgs = [ct.rgs['rg3'], ct.rgs['rg4']]
        # Get destination-replication-group for partition: p1
        rg_dest = ct._elect_dest_replication_group(
            rg_source.count_replica(p1),
            under_replicated_rgs,
            p1,
        )

        # Dest-replica can be either 'rg3' or 'rg4' with replica-count 1
        assert rg_dest.id in ['rg3', 'rg4']

        # Case 2: rg-source == 'rg2': No destination group found
        rg_source = ct.rgs['rg2']
        # Get destination-replication-group for partition: p1
        rg_dest = ct._elect_dest_replication_group(
            rg_source.count_replica(p1),
            under_replicated_rgs,
            p1,
        )

        # Since none of under-replicated-groups (rg3, and rg4) have lower
        # 2-1=0 replicas for the given partition p1
        # No eligible dest-group is there where partition can be sent to
        assert rg_dest is None

    def test_rebalance_partition_imbalanced_case1(self):
        # Test imbalanced partitions for below cases
        # Note: In initial-assignment, all partitions with id-1 are 'imbalanced'
        ct = self.build_cluster_topology()
        # CASE 1: repl-factor % rg-count == 0
        # (1a): repl-factor == rg-count
        # p1: replicas: ('T0', 1): [2,3]
        p1 = ct.partitions[('T0', 1)]
        # rg-imbalanced p1
        opt_cnt = 1    # 2/2
        ct._rebalance_partition(p1)

        # Verify partition is rg-balanced
        self.assert_rg_balanced_partition(ct, p1, opt_cnt)

        # (1b):  repl-count % rg-count == 0 and repl-count > rg-count
        # p1: replicas: ('T1',1): [0,1,2,4]
        p1 = ct.partitions[('T1', 1)]
        # Assert originally-imbalanced p1
        opt_cnt = 2    # 4/2
        ct._rebalance_partition(p1)

        # Verify partition is rg-balanced
        self.assert_rg_balanced_partition(ct, p1, opt_cnt)

    def test_rebalance_partition_imbalanced_case2(self):
        ct = self.build_cluster_topology()
        # CASE 2: repl-factor % rg-count > 0
        # p1: replicas ('T3', 1): [0,1,4]
        p1 = ct.partitions[('T3', 1)]
        # rg-imbalanced p1
        opt_cnt = 1    # 3/2
        extra_cnt = 1  # 3%2
        ct._rebalance_partition(p1)

        # Verify partition is now rg-balanced
        self.assert_rg_balanced_partition(ct, p1, opt_cnt, extra_cnt)

    def test_rebalance_partition_balanced(self):
        # Test already balanced partitions in given example for different cases
        # Analyze Cases 1a, 1b
        ct = self.build_cluster_topology()
        # CASE 1: repl-factor % rg-count == 0
        # (1a): repl-factor == rg-count
        # p1: replicas: ('T0', 0): [1,2]
        p1 = ct.partitions[('T0', 0)]
        opt_cnt = 1    # 2/2
        self.assert_rg_balanced_partition(ct, p1, opt_cnt)
        ct._rebalance_partition(p1)

        # Verify no change in replicas after rebalancing
        self.rg_rebalance_assert_no_change(ct, p1)

        # (1b):  repl-count % rg-count == 0 and repl-count > rg-count
        # p1: replicas: ('T1',0): [0,1,2,3]
        p1 = ct.partitions[('T1', 0)]
        opt_cnt = 2    # 4/2
        self.assert_rg_balanced_partition(ct, p1, opt_cnt)

        # Verify no change in replicas after rebalancing
        self.rg_rebalance_assert_no_change(ct, p1)

    def test_rebalance_partition_balanced_case2(self):
        # Test already balanced partitions in given example for different cases
        # Analyze Cases 2a, 2b
        ct = self.build_cluster_topology()
        # CASE 2: repl-factor % rg-count > 0
        # (2a): repl-factor < rg-count
        # p1: replicas ('T2', 0): [2]
        p1 = ct.partitions[('T2', 0)]
        opt_cnt = 0    # 1/2
        extra_cnt = 1  # 1%2
        self.assert_rg_balanced_partition(ct, p1, opt_cnt, extra_cnt)

        # Verify no change in replicas after rebalancing
        self.rg_rebalance_assert_no_change(ct, p1)

        # (2b): repl-factor > rg-count
        # p1: replicas: ('T3', 0), [0,1,2]): ['rg1', 'rg1', 'rg2']
        p1 = ct.partitions[('T3', 0)]
        opt_cnt = 1    # 3/2
        extra_cnt = 1  # 3%2
        self.assert_rg_balanced_partition(ct, p1, opt_cnt, extra_cnt)

        # Verify no change in replicas after rebalancing
        self.rg_rebalance_assert_no_change(ct, p1)

    def rg_rebalance_assert_no_change(self, ct, p1):
        """Verifies that there are no replica change after rebalancing."""
        old_replicas = p1.replicas
        ct._rebalance_partition(p1)

        # Verify no replica change for partition
        assert old_replicas == p1.replicas

    def assert_rg_balanced_partition(self, ct, p1, opt_cnt, extra_cnt=0):
        for rg in ct.rgs.itervalues():
            replica_cnt_rg = rg.count_replica(p1)

            # Verify for evenly-balanced partition p1
            assert replica_cnt_rg == opt_cnt or\
                replica_cnt_rg == opt_cnt + extra_cnt

    def srange(self, n):
        """Return list of integers as string from 0 to n-1."""
        return {str(x): {"host": "host%s" % x} for x in range(n)}

    def assert_valid(self, new_assignment, orig_assignment, orig_brokers):
        """Assert if new-assignment is valid based on given assignment.

        Asserts the results for following parameters:
        a) Asserts that keys in both assignments are same
        b) Asserts that replication-factor of result remains same
        c) Assert that new-replica-brokers are amongst given broker-list
        """

        # Verify that partitions remain same
        assert sorted(orig_assignment.keys()) == sorted(new_assignment.keys())
        for t_p, new_replicas in new_assignment.iteritems():
            orig_replicas = orig_assignment[t_p]
            # Verify that new-replicas are amongst given broker-list
            assert all([broker in orig_brokers for broker in new_replicas])
            # Verify that replication-factor remains same
            assert len(new_replicas) == len(orig_replicas)

    # Tests for leader-balancing
    def test_rebalance_leaders_balanced_case1(self):
        # Already balanced-assignment with evenly-distributed
        # (broker-id: leader-count): {0: 1, 1:1, 2:1}
        # opt-count: 3/3 = 1, extra-count: 3%3 = 0
        assignment = dict(
            [
                ((u'T0', 0), ['1', '2']),
                ((u'T0', 1), ['2', '0']),
                ((u'T1', 0), ['0', '2']),
            ]
        )
        ct = self.build_cluster_topology(assignment, self.srange(3))
        orig_assignment = ct.assignment
        ct.rebalance_leaders()
        _, net_imbal, _ = get_leader_imbalance_stats(ct.brokers.values())

        # No changed in already-balanced assignment
        assert orig_assignment == ct.assignment
        # Assert leader-balanced
        assert net_imbal == 0

    def test_rebalance_leaders_balanced_case2(self):
        # Already balanced-assignment NOT evenly-distributed
        # (broker-id: leader-count): {0: 1, 1:1, 2:1}
        # opt-count: 2/3 = 0, extra-count: 2%3 = 2
        assignment = dict(
            [
                ((u'T0', 0), ['1', '2']),
                ((u'T0', 1), ['2', '0']),
            ]
        )
        ct = self.build_cluster_topology(assignment, self.srange(3))
        orig_assignment = ct.assignment
        ct.rebalance_leaders()
        _, net_imbal, _ = get_leader_imbalance_stats(ct.brokers.values())

        # No changed in already-balanced assignment
        assert orig_assignment == ct.assignment
        # Assert leader-balanced
        assert net_imbal == 0

    def test_rebalance_leaders_unbalanced_case1(self):
        # Balance leader-imbalance successfully
        # (broker-id: leader-count): {0: 0, 1:2, 2:1}
        # Net-leader-imbalance: 1
        # opt-count: 3/3 = 1, extra-count: 3%3 = 0
        assignment = dict(
            [
                ((u'T0', 0), ['1', '2']),
                ((u'T0', 1), ['2', '0']),
                ((u'T1', 0), ['1', '0']),
            ]
        )
        ct = self.build_cluster_topology(assignment, self.srange(3))
        orig_assignment = ct.assignment
        ct.rebalance_leaders()

        # Verify if valid-leader assignment
        self.assert_leader_valid(orig_assignment, ct.assignment)
        # New-leader imbalance-count be less than previous imbal count
        _, new_leader_imbal, new_leaders_per_broker = \
            get_leader_imbalance_stats(ct.brokers.values())
        # Verify leader-balanced
        assert new_leader_imbal == 0
        # Verify partitions-changed assignment
        assert new_leaders_per_broker['0'] == 1
        assert new_leaders_per_broker['1'] == 1
        assert new_leaders_per_broker['2'] == 1

    def test_rebalance_leaders_unbalanced_case2(self):
        # (Broker: leader-count): {0: 2, 1: 1, 2:0}
        # opt-count: 3/3 = 1, extra-count = 0
        # Leader-imbalance-value: 1
        assignment = dict(
            [
                ((u'T0', 0), ['1', '2']),
                ((u'T1', 1), ['0', '1']),
                ((u'T1', 0), ['0']),
            ]
        )
        ct = self.build_cluster_topology(assignment, self.srange(3))
        ct.rebalance_leaders()

        # Verify leader-balanced
        _, leader_imbal, _ = get_leader_imbalance_stats(ct.brokers.values())
        assert leader_imbal == 0

    def test_rebalance_leaders_unbalanced_case2a(self):
        # (Broker: leader-count): {0: 2, 1: 1, 2:0, 3:1}
        # opt-count: 3/4 = 1, extra-count = 3
        # Leader-imbalance-value: 1
        # imbalanced-broker: 0,2; balanced-brokers: 1,3
        assignment = dict(
            [
                ((u'T0', 0), ['3', '2']),
                ((u'T0', 1), ['1', '3']),
                ((u'T1', 1), ['0', '1']),
                ((u'T1', 0), ['0']),
            ]
        )
        ct = self.build_cluster_topology(assignment, self.srange(4))
        ct.rebalance_leaders()

        # Verify balanced
        _, leader_imbal, _ = get_leader_imbalance_stats(ct.brokers.values())
        assert leader_imbal == 0
        # Verify that (T0, 1) also swapped even if 1 and 3 were balanced
        # Rebalancing through non-followers
        replica_ids = [b.id for b in ct.partitions[('T0', 1)].replicas]
        assert replica_ids == ['3', '1']

    def test_rebalance_leaders_unbalanced_case2b(self):
        assignment = dict(
            [
                ((u'T0', 0), ['3', '2']),
                ((u'T1', 0), ['1', '2']),
                ((u'T1', 1), ['0', '1']),
                ((u'T2', 0), ['0']),
            ]
        )
        ct = self.build_cluster_topology(assignment, self.srange(4))
        ct.rebalance_leaders()

        # Verify leader-balanced
        _, leader_imbal, _ = get_leader_imbalance_stats(ct.brokers.values())
        assert leader_imbal == 0

    def test_rebalance_leaders_unbalanced_case2c(self):
        # Broker-2 imbalance value: 2 with different brokers
        # Broker-2 requests leadership from multiple brokers (0, 1) once
        assignment = dict(
            [
                ((u'T1', 0), ['1', '2']),
                ((u'T1', 1), ['0', '1']),
                ((u'T2', 0), ['0']),
                ((u'T2', 1), ['0']),
                ((u'T3', 0), ['3', '2']),
                ((u'T3', 1), ['1', '3']),
                ((u'T4', 0), ['1']),
                ((u'T4', 2), ['3']),
            ]
        )
        ct = self.build_cluster_topology(assignment, self.srange(4))
        ct.rebalance_leaders()

        # Verify leader-balanced
        _, leader_imbal, _ = get_leader_imbalance_stats(ct.brokers.values())
        assert leader_imbal == 0

    def test_rebalance_leaders_unbalanced_case2d(self):
        # Broker-2 imbalanced with same brokers
        # Broker-2 requests leadership from same broker-1 twice
        assignment = dict(
            [
                ((u'T1', 0), ['1', '2']),
                ((u'T1', 1), ['0', '1']),
                ((u'T1', 2), ['0']),
                ((u'T1', 3), ['1', '2']),
                ((u'T1', 4), ['0', '1']),
                ((u'T1', 5), ['0']),
            ]
        )
        ct = self.build_cluster_topology(assignment, self.srange(3))
        ct.rebalance_leaders()

        # Verify leader-balanced
        _, leader_imbal, _ = get_leader_imbalance_stats(ct.brokers.values())
        assert leader_imbal == 0

    def test_rebalance_leaders_unbalanced_case2e(self):
        # Imbalance-val 2
        # Multiple imbalanced brokers (2, 5) gets non-follower balanced
        # from multiple brokers (1,4)
        assignment = dict(
            [
                ((u'T1', 0), ['1', '2']),
                ((u'T1', 1), ['0', '1']),
                ((u'T2', 0), ['0']),
                ((u'T3', 0), ['4', '5']),
                ((u'T3', 1), ['3', '4']),
                ((u'T4', 0), ['3']),
            ]
        )
        ct = self.build_cluster_topology(assignment, self.srange(6))
        ct.rebalance_leaders()

        # Verify leader-balanced
        _, leader_imbal, _ = get_leader_imbalance_stats(ct.brokers.values())
        assert leader_imbal == 0

    def test_rebalance_leaders_unbalanced_case3(self):
        # Imbalanced 0 and 2. No re-balance possible.
        assignment = dict(
            [
                ((u'T1', 0), ['1', '2']),
                ((u'T1', 1), ['0']),
                ((u'T2', 0), ['0']),
            ]
        )
        ct = self.build_cluster_topology(assignment, self.srange(3))
        ct.rebalance_leaders()

        # Verify still leader-imbalanced
        _, leader_imbal, _ = get_leader_imbalance_stats(ct.brokers.values())
        assert leader_imbal == 1
        # No change in assignment
        assert sorted(ct.assignment) == sorted(assignment)

    def test_rebalance_leaders_unbalanced_case4(self):
        # Imbalanced assignment
        # Partial leader-imbalance possible
        # (Broker: leader-count): {0: 3, 1: 1, 2:0}
        # opt-count: 5/3 = 1, extra-count = 2
        assignment = dict(
            [
                ((u'T0', 0), ['1', '2']),
                ((u'T0', 1), ['0', '2']),
                ((u'T1', 0), ['0']),
                ((u'T1', 1), ['0']),
                ((u'T1', 2), ['0']),
            ]
        )

        ct = self.build_cluster_topology(assignment, self.srange(3))
        _, net_imbal, _ = get_leader_imbalance_stats(ct.brokers.values())
        ct.rebalance_leaders()
        _, new_net_imbal, new_leaders_per_broker = get_leader_imbalance_stats(
            ct.brokers.values(),
        )
        # Verify that net-imbalance has reduced but not zero
        assert new_net_imbal > 0 and new_net_imbal < net_imbal
        # Verify the changes in leaders-per-broker count
        assert new_leaders_per_broker['2'] == 1
        assert new_leaders_per_broker['1'] == 1
        assert new_leaders_per_broker['0'] == 3

    def test_rebalance_leaders_unbalanced_case2f(self):
        assignment = dict(
            [
                ((u'T0', 0), ['2', '0']),
                ((u'T1', 0), ['2', '0']),
                ((u'T1', 1), ['0']),
                ((u'T2', 0), ['1']),
                ((u'T2', 1), ['2']),
            ]
        )
        ct = self.build_cluster_topology(assignment, self.srange(3))
        ct.rebalance_leaders()

        # Verify leader-balanced
        _, leader_imbal, _ = get_leader_imbalance_stats(ct.brokers.values())
        assert leader_imbal == 0

    def test_rebalance_leaders_unbalanced_case5(self):
        # Special case, wherein none under-balanced
        # but 0 is overbalanced
        assignment = dict(
            [
                ((u'T1', 1), ['0', '1']),
                ((u'T2', 0), ['0']),
                ((u'T2', 1), ['0']),
                ((u'T3', 0), ['2', '3']),
                ((u'T3', 1), ['3', '1']),
                ((u'T4', 0), ['1']),
            ]
        )
        ct = self.build_cluster_topology(assignment, self.srange(4))
        ct.rebalance_leaders()

        # Verify leader-balanced
        _, leader_imbal, _ = get_leader_imbalance_stats(ct.brokers.values())
        assert leader_imbal == 0

    def test__rebalance_groups_partition_cnt_case1(self):
        # rg1 has 6 partitions
        # rg2 has 2 partitions
        # Both rg's are balanced(based on replica-count) initially
        # Result: rg's will be balanced for partition-count
        assignment = dict(
            [
                ((u'T1', 1), ['0', '1', '2']),
                ((u'T1', 0), ['1']),
                ((u'T3', 0), ['1']),
                ((u'T2', 0), ['0', '1', '3']),
            ]
        )
        ct = self.build_cluster_topology(assignment, self.srange(4))
        # Re-balance replication-groups for partition-count
        ct._rebalance_groups_partition_cnt()

        # Verify both replication-groups have same partition-count
        assert len(ct.rgs['rg1'].partitions) == len(ct.rgs['rg2'].partitions)
        _, total_movements = \
            calculate_partition_movement(assignment, ct.assignment)
        # Verify minimum partition movements 2
        assert total_movements == 2
        net_imbal, _ = get_replication_group_imbalance_stats(
            ct.rgs.values(),
            ct.partitions.values(),
        )
        # Verify replica-count imbalance remains unaltered
        assert net_imbal == 0

    def test__rebalance_groups_partition_cnt_case2(self):
        # 1 over-balanced, 2 under-balanced replication-groups
        # rg1 has 4 partitions
        # rg2 has 1 partition
        # rg3 has 1 partition
        # All rg's are balanced(based on replica-count) initially
        # Result: rg's will be balanced for partition-count
        assignment = dict(
            [
                ((u'T1', 1), ['0', '2']),
                ((u'T3', 1), ['0']),
                ((u'T3', 0), ['0']),
                ((u'T2', 0), ['0', '5']),
            ]
        )
        brokers = {
            '0': mock.MagicMock(),
            '2': mock.MagicMock(),
            '5': mock.MagicMock(),
        }
        ct = self.build_cluster_topology(assignment, brokers)
        # Re-balance brokers
        ct._rebalance_groups_partition_cnt()

        # Verify all replication-groups have same partition-count
        assert len(ct.rgs['rg1'].partitions) == len(ct.rgs['rg2'].partitions)
        assert len(ct.rgs['rg1'].partitions) == len(ct.rgs['rg3'].partitions)
        _, total_movements = \
            calculate_partition_movement(assignment, ct.assignment)
        # Verify minimum partition movements 2
        assert total_movements == 2
        net_imbal, _ = get_replication_group_imbalance_stats(
            ct.rgs.values(),
            ct.partitions.values(),
        )
        # Verify replica-count imbalance remains 0
        assert net_imbal == 0

    def test__rebalance_groups_partition_cnt_case3(self):
        # 1 over-balanced, 1 under-balanced, 1 opt-balanced replication-group
        # rg1 has 3 partitions
        # rg2 has 2 partitions
        # rg3 has 1 partition
        # All rg's are balanced(based on replica-count) initially
        # Result: rg's will be balanced for partition-count
        assignment = dict(
            [
                ((u'T1', 1), ['0', '2']),
                ((u'T3', 1), ['2']),
                ((u'T3', 0), ['0']),
                ((u'T2', 0), ['0', '5']),
            ]
        )
        brokers = {
            '0': mock.MagicMock(),
            '2': mock.MagicMock(),
            '5': mock.MagicMock(),
        }
        ct = self.build_cluster_topology(assignment, brokers)
        # Re-balance brokers across replication-groups
        ct._rebalance_groups_partition_cnt()

        # Verify all replication-groups have same partition-count
        assert len(ct.rgs['rg1'].partitions) == len(ct.rgs['rg2'].partitions)
        assert len(ct.rgs['rg1'].partitions) == len(ct.rgs['rg3'].partitions)
        _, total_movements = \
            calculate_partition_movement(assignment, ct.assignment)
        # Verify minimum partition movements
        assert total_movements == 1
        net_imbal, _ = get_replication_group_imbalance_stats(
            ct.rgs.values(),
            ct.partitions.values(),
        )
        # Verify replica-count imbalance remains 0
        assert net_imbal == 0

    def test_update_cluster_topology_invalid_broker(self):
        assignment = dict([((u'T0', 0), ['1', '2'])])
        new_assignment = dict([((u'T0', 0), ['1', '3'])])
        ct = self.build_cluster_topology(assignment, self.srange(3))

        with pytest.raises(InvalidBrokerIdError):
            ct.update_cluster_topology(new_assignment)

    def test_update_cluster_topology_invalid_partition(self):
        assignment = dict([((u'T0', 0), ['1', '2'])])
        new_assignment = dict([((u'invalid_topic', 0), ['1', '0'])])
        ct = self.build_cluster_topology(assignment, self.srange(3))

        with pytest.raises(InvalidPartitionError):
            ct.update_cluster_topology(new_assignment)

    def test_update_cluster_topology(self):
        assignment = dict(
            [
                ((u'T0', 0), ['1', '2']),
                ((u'T0', 1), ['2', '0']),
                ((u'T1', 0), ['0', '2']),
            ]
        )
        new_assignment = dict(
            [
                ((u'T0', 0), ['1', '2']),
                ((u'T0', 1), ['1', '2']),
            ]
        )
        ct = self.build_cluster_topology(assignment, self.srange(3))

        ct.update_cluster_topology(new_assignment)

        # Verify updates of partition and broker objects
        r_T0_0 = [b.id for b in ct.partitions[(u'T0', 0)].replicas]
        r_T0_1 = [b.id for b in ct.partitions[(u'T0', 1)].replicas]
        r_T1_0 = [b.id for b in ct.partitions[(u'T1', 0)].replicas]
        assert r_T0_0 == ['1', '2']
        assert r_T0_1 == ['1', '2']
        assert r_T1_0 == ['0', '2']

        # Assert partitions of brokers get updated
        assert ct.brokers['0'].partitions == set([
            ct.partitions[(u'T1', 0)],
        ])

        assert ct.brokers['1'].partitions == set([
            ct.partitions[(u'T0', 0)],
            ct.partitions[(u'T0', 1)],
        ])

        assert ct.brokers['2'].partitions == set([
            ct.partitions[(u'T0', 0)],
            ct.partitions[(u'T0', 1)],
            ct.partitions[(u'T1', 0)],
        ])

    def assert_leader_valid(self, orig_assignment, new_assignment):
        """Verify that new-assignment complies with just leader changes.

        Following characteristics are verified for just leader-changes.
        a) partitions remain same
        b) replica set remains same
        """
        # Partition-list remains unchanged
        assert sorted(orig_assignment.keys()) == sorted(new_assignment.keys())
        # Replica-set remains same
        for partition, orig_replicas in orig_assignment.iteritems():
            set(orig_replicas) == set(new_assignment[partition])

    def test_replace_broker_leader(self):
        assignment = dict(
            [
                ((u'T1', 0), ['1', '2']),
                ((u'T1', 1), ['0', '2']),
                ((u'T2', 0), ['1']),
            ]
        )
        ct = self.build_cluster_topology(assignment, self.srange(4))
        ct.replace_broker('1', '3')

        assert ct.brokers['1'].partitions == set([])
        assert ct.brokers['3'].partitions == set([
            ct.partitions[(u'T1', 0)],
            ct.partitions[(u'T2', 0)],
        ])
        assert ct.partitions[(u'T1', 0)].replicas == [
            ct.brokers['3'],
            ct.brokers['2'],
        ]

    def test_replace_broker_non_leader(self):
        assignment = dict(
            [
                ((u'T1', 0), ['1', '2']),
                ((u'T2', 0), ['1']),
            ]
        )
        ct = self.build_cluster_topology(assignment, self.srange(4))
        ct.replace_broker('2', '0')

        assert ct.brokers['2'].partitions == set([])
        assert ct.brokers['0'].partitions == set([ct.partitions[(u'T1', 0)]])
        assert ct.partitions[(u'T1', 0)].replicas == [
            ct.brokers['1'],
            ct.brokers['0'],
        ]

    def test_replace_broker_invalid_source_broker(self):
        assignment = dict([((u'T1', 0), ['0', '1'])])
        ct = self.build_cluster_topology(assignment, self.srange(3))

        with pytest.raises(InvalidBrokerIdError):
            ct.replace_broker('444', '2')

    def test_replace_broker_invalid_destination_broker(self):
        assignment = dict([((u'T1', 0), ['0', '1'])])
        ct = self.build_cluster_topology(assignment, self.srange(3))

        with pytest.raises(InvalidBrokerIdError):
            ct.replace_broker('0', '444')



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import sys

from .command import ClusterManagerCmd
from kafka_utils.util.validation import assignment_to_plan
from kafka_utils.util.validation import validate_plan


DEFAULT_MAX_PARTITION_MOVEMENTS = 1
DEFAULT_MAX_LEADER_CHANGES = 5


class RebalanceCmd(ClusterManagerCmd):

    def __init__(self):
        super(RebalanceCmd, self).__init__()
        self.log = logging.getLogger('ClusterRebalance')

    def build_subparser(self, subparsers):
        subparser = subparsers.add_parser(
            'rebalance',
            description='Rebalance cluster by moving partitions across brokers '
            'and changing the preferred replica.',
            help='This command is used to rebalance a Kafka cluster. Based on '
            'the given flags this tool will generate and submit a reassinment '
            'plan that will evenly distribute partitions and leaders '
            'across the brokers of the cluster. The replication groups option '
            'moves the replicas of the same partition to separate replication '
            'making the cluster resilient to the failure of one of more zones.'
        )
        subparser.add_argument(
            '--replication-groups',
            action='store_true',
            help='Evenly distributes replicas over replication-groups.',
        )
        subparser.add_argument(
            '--brokers',
            action='store_true',
            help='Evenly distributes partitions optimally over brokers'
            ' with minimal movements for each replication-group.',
        )
        subparser.add_argument(
            '--leaders',
            action='store_true',
            help='Evenly distributes leaders optimally over brokers.',
        )
        subparser.add_argument(
            '--max-partition-movements',
            type=self.positive_int,
            default=DEFAULT_MAX_PARTITION_MOVEMENTS,
            help='Maximum number of partition-movements in final set of actions.'
                 ' DEFAULT: %(default)s. RECOMMENDATION: Should be at least max '
                 'replication-factor across the cluster.',
        )
        subparser.add_argument(
            '--max-leader-changes',
            type=self.positive_int,
            default=DEFAULT_MAX_LEADER_CHANGES,
            help='Maximum number of actions with leader-only changes.'
                 ' DEFAULT: %(default)s',
        )
        return subparser

    def run_command(self, ct):
        """Get executable proposed plan(if any) for display or execution."""
        base_assignment = ct.assignment
        assignment = self.build_balanced_assignment(ct)

        if not validate_plan(
            assignment_to_plan(assignment),
            assignment_to_plan(base_assignment),
        ):
            self.log.error('Invalid latest-cluster assignment. Exiting.')
            sys.exit(1)

        # Reduce the proposed assignment based on max_partition_movements
        # and max_leader_changes
        reduced_assignment = self.get_reduced_assignment(
            base_assignment,
            assignment,
            self.args.max_partition_movements,
            self.args.max_leader_changes,
        )
        if reduced_assignment:
            self.process_assignment(reduced_assignment)
        else:
            self.log.info("Cluster already balanced. No actions to perform.")

    def build_balanced_assignment(self, ct):
        # Balancing to be done in the given order only
        # Rebalance replication-groups
        if self.args.replication_groups:
            self.log.info(
                'Re-balancing replica-count over replication groups: %s',
                ', '.join(ct.rgs.keys()),
            )
            ct.rebalance_replication_groups()

        # Rebalance broker-partition count per replication-groups
        if self.args.brokers:
            self.log.info(
                'Re-balancing partition-count across brokers: %s',
                ', '.join(str(e) for e in ct.brokers.keys()),
            )
            ct.rebalance_brokers()

        # Rebalance broker as leader count per broker
        if self.args.leaders:
            self.log.info(
                'Re-balancing leader-count across brokers: %s',
                ', '.join(str(e) for e in ct.brokers.keys()),
            )
            ct.rebalance_leaders()
        return ct.assignment



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from mock import Mock
from mock import sentinel
from pytest import fixture

from .helper import create_and_attach_partition
from .helper import create_broker
from kafka_utils.kafka_cluster_manager.cluster_info.broker import Broker
from kafka_utils.kafka_cluster_manager.cluster_info.partition import Partition
from kafka_utils.kafka_cluster_manager.cluster_info.topic import Topic


class TestBroker(object):

    @fixture
    def partition(self):
        mock_topic = Mock(spec=Topic, id='t1')
        return Partition(mock_topic, 0)

    # Initial broker-set empty
    def test_partitions_empty(self):
        broker = Broker('test-broker')

        assert broker.partitions == set()

    def test_partitions(self):
        broker = Broker('test-broker', partitions=set([sentinel.p1, sentinel.p2]))

        assert broker.partitions == set([sentinel.p1, sentinel.p2])

    def test_remove_partition(self, partition):
        p1 = partition
        b1 = Broker('test-broker', partitions=set([p1]))
        p1.add_replica(b1)

        # Remove partition
        b1.remove_partition(p1)

        assert p1 not in b1.partitions
        assert b1 not in p1.replicas

    def test_add_partition(self, create_partition):
        p10 = create_partition('t1', 0)
        p20 = create_partition('t2', 0)
        broker = Broker('test-broker', partitions=set([p10]))

        broker.add_partition(p20)

        assert broker.partitions == set([p10, p20])
        assert p20.replicas == [broker]

    def test_topics(self):
        partitions = set([
            Mock(spec=Partition, topic=sentinel.t1),
            Mock(spec=Partition, topic=sentinel.t1),
            Mock(spec=Partition, topic=sentinel.t2),
        ])
        broker = Broker('test-broker', partitions=partitions)

        assert broker.topics == set([sentinel.t1, sentinel.t2])

    def test_count_partition(self):
        t1 = sentinel.t1
        partitions = set([
            Mock(spec=Partition, topic=t1, replicas=sentinel.r1),
            Mock(spec=Partition, topic=t1, replicas=sentinel.r1),
            Mock(spec=Partition, topic=sentinel.t2, replicas=sentinel.r1),
        ])
        broker = Broker('test-broker', partitions=partitions)

        assert broker.count_partitions(t1) == 2
        assert broker.count_partitions(sentinel.t3) == 0

    def test_move_partition(self, partition):
        victim_partition = partition
        source_broker = Broker('b1', partitions=set([victim_partition]))
        victim_partition.add_replica(source_broker)
        dest_broker = Broker('b2')

        # Move partition
        source_broker.move_partition(victim_partition, dest_broker)

        # Verify partition moved from source to destination broker
        assert victim_partition not in source_broker.partitions
        assert victim_partition in dest_broker.partitions
        # Verify that victim-partition-replicas has replaced
        # source-broker to dest-broker
        assert source_broker not in victim_partition.replicas
        assert dest_broker in victim_partition.replicas

    def test_count_preferred_replica(self, partition):
        p1 = partition
        b1 = Broker('test-broker', partitions=set([p1]))
        p1.add_replica(b1)
        p2 = Mock(spec=Partition, topic=sentinel.t1, replicas=[sentinel.b2])
        b1.add_partition(p2)

        assert b1.count_preferred_replica() == 1

    def test_get_preferred_partition(self):
        t1 = Topic('t1', 1)
        t2 = Topic('t2', 1)
        t3 = Topic('t3', 1)
        # Naming convention: p10 -> topic t1 partition 0
        p10 = create_and_attach_partition(t1, 0)
        p11 = create_and_attach_partition(t1, 1)
        p20 = create_and_attach_partition(t2, 0)
        p21 = create_and_attach_partition(t2, 1)
        p30 = create_and_attach_partition(t3, 0)
        source = create_broker('b1', [p10, p11, p20])
        dest = create_broker('b2', [p21, p30])
        sibling_distance = {t1: -2, t2: 0, t3: 1}

        actual = source.get_preferred_partition(dest, sibling_distance)

        assert actual in (p10, p11)

    def test_get_preferred_partition_no_preferred(self):
        t1 = Topic('t1', replication_factor=2)
        p10 = create_and_attach_partition(t1, 0)
        p11 = create_and_attach_partition(t1, 1)
        source = create_broker('b1', [p10, p11])
        dest = create_broker('b2', [p10, p11])
        sibling_distance = {t1: 0}

        actual = source.get_preferred_partition(dest, sibling_distance)

        assert actual is None

    def test_mark_decommissioned(self):
        broker = Broker('test-broker')
        broker.mark_decommissioned()

        assert broker.decommissioned

    def test_empty(self):
        broker = Broker('test-broker')
        # No partitions for this broker
        assert broker.empty()

    def test_not_empty(self, partition):
        broker = Broker('test-broker', partitions=set([partition]))

        assert not broker.empty()

    def test_mark_inactive(self):
        broker = Broker('test-broker')
        broker.mark_inactive()

        assert broker.inactive



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import functools

from kafka import KafkaClient

from kafka_utils.util.protocol import KafkaToolProtocol


class KafkaToolClient(KafkaClient):
    '''
    Extends the KafkaClient class, and includes a method for sending offset
    commit requests to Kafka.
    '''

    def send_offset_commit_request_kafka(
            self, group, payloads=[],
            fail_on_error=True, callback=None):
        encoder = functools.partial(
            KafkaToolProtocol.encode_offset_commit_request_kafka,
            group=group,
        )
        decoder = KafkaToolProtocol.decode_offset_commit_response
        resps = self._send_consumer_aware_request(group, payloads, encoder, decoder)

        return [resp if not callback else callback(resp) for resp in resps
                if not fail_on_error or not self._raise_on_response_error(resp)]



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import sys

from .offset_manager import OffsetManagerBase
from kafka_utils.util import print_json
from kafka_utils.util.client import KafkaToolClient
from kafka_utils.util.monitoring import get_consumer_offsets_metadata


class OffsetGet(OffsetManagerBase):

    @classmethod
    def setup_subparser(cls, subparsers):
        parser_offset_get = subparsers.add_parser(
            "offset_get",
            description="Get consumer offsets for the"
            " specified consumer group",
            add_help=False
        )
        parser_offset_get.add_argument(
            "-h", "--help", action="help",
            help="Show this help message and exit."
        )
        parser_offset_get.add_argument(
            'groupid',
            help="Consumer Group ID whose offsets shall be fetched."
        )
        parser_offset_get.add_argument(
            "--topic",
            help="Kafka topic whose offsets shall be fetched. If no topic is "
            "specified, offsets from all topics that the consumer is "
            "subscribed to, shall be fetched."
        )
        parser_offset_get.add_argument(
            "--partitions", nargs='+', type=int,
            help="List of partitions within the topic. If no partitions are "
            "specified, offsets from all partitions of the topic shall "
            "be fetched."
        )
        parser_offset_get.add_argument(
            "--watermark",
            choices=["high", "low", "current", "all"],
            help="Type of offset watermark. \"high\" represents the offset "
            "corresponding to the latest message. \"low\" represents "
            "the offset corresponding to the earliest message. \"current\" "
            "represents the current commited offset for this consumer",
            default="all"
        )
        parser_offset_get.add_argument(
            '--storage', choices=['zookeeper', 'kafka', 'dual'],
            help="String describing where to fetch the committed offsets.",
            default='dual'
        )
        parser_offset_get.add_argument(
            "-j", "--json", action="store_true",
            help="Export data in json format."
        )
        parser_offset_get.set_defaults(command=cls.run)

    @classmethod
    def run(cls, args, cluster_config):
        # Setup the Kafka client
        client = KafkaToolClient(cluster_config.broker_list)
        client.load_metadata_for_topics()

        topics_dict = cls.preprocess_args(
            args.groupid, args.topic, args.partitions, cluster_config, client
        )

        consumer_offsets_metadata = cls.get_offsets(
            client,
            args.groupid,
            topics_dict,
            args.storage,
        )

        # Warn the user if a topic being subscribed to does not exist in
        # Kafka.
        for topic in topics_dict:
            if topic not in consumer_offsets_metadata:
                print(
                    "Warning: Topic {topic} or one or more of it's partitions "
                    "do not exist in Kafka".format(topic=topic),
                    file=sys.stderr,
                )
        client.close()
        if args.json:
            print_json(consumer_offsets_metadata)
        else:
            cls.print_output(consumer_offsets_metadata, args.watermark)

    @classmethod
    def get_offsets(cls, client, group, topics_dict, storage):
        try:
            return get_consumer_offsets_metadata(
                client, group, topics_dict, False, storage,
            )
        except:
            print(
                "Error: Encountered error with Kafka, please try again later.",
                file=sys.stderr
            )
            raise

    @classmethod
    def print_output(cls, consumer_offsets_metadata, watermark_filter):
        for topic, metadata_tuples in consumer_offsets_metadata.iteritems():
            print ("Topic Name: {topic}".format(topic=topic))
            for metadata_tuple in metadata_tuples:
                print (
                    "\tPartition ID: {partition}".format(
                        partition=metadata_tuple.partition
                    )
                )
                if watermark_filter == "all" or watermark_filter == "high":
                    print(
                        "\t\tHigh Watermark: {high}".format(
                            high=metadata_tuple.highmark
                        )
                    )
                if watermark_filter == "all" or watermark_filter == "low":
                    print(
                        "\t\tLow Watermark: {low}".format(
                            low=metadata_tuple.lowmark
                        )
                    )
                if watermark_filter == "all" or watermark_filter == "current":
                    print(
                        "\t\tCurrent Offset: {current}".format(
                            current=metadata_tuple.current
                        )
                    )



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import contextlib
import sys

import mock
import pytest

from kafka_utils.kafka_consumer_manager. \
    commands.offset_manager import OffsetManagerBase
from kafka_utils.util.client import KafkaToolClient


class TestOffsetManagerBase(object):
    topics_partitions = {
        "topic1": [0, 1, 2],
        "topic2": [0, 1, 2, 3],
        "topic3": [0, 1],
    }

    def _get_partition_ids_for_topic(self, topic):
        try:
            return self.topics_partitions[topic]
        # Since we patch sys.exit, let's mask this exception since this call
        # is triggered after the call to sys.exit
        except KeyError:
            pass

    @pytest.fixture
    def mock_kafka_client(self):
        mock_kafka_client = mock.MagicMock(
            spec=KafkaToolClient
        )
        mock_kafka_client.get_partition_ids_for_topic. \
            side_effect = self._get_partition_ids_for_topic
        return mock_kafka_client

    @contextlib.contextmanager
    def mock_get_topics(self):
        with mock.patch.object(
            OffsetManagerBase,
            "get_topics_from_consumer_group_id",
            spec=OffsetManagerBase.get_topics_from_consumer_group_id,
            return_value=["topic1", "topic2", "topic3"],
        ) as mock_get_topics:
            yield mock_get_topics

    def test_preprocess_args_with_topics_and_partitions(self, mock_kafka_client):
        args = mock.Mock(
            groupid="some_group",
            topic="topic1",
            partitions=[0, 1]
        )
        expected_topics_dict = {"topic1": [0, 1]}
        with self.mock_get_topics() as mock_get_topics:
            topics_dict = OffsetManagerBase.preprocess_args(
                args.groupid,
                args.topic,
                args.partitions,
                mock.Mock(),
                mock_kafka_client
            )
            assert mock_get_topics.called
            assert topics_dict == expected_topics_dict

    def test_preprocess_args_with_topics_no_partitions(self, mock_kafka_client):
        args = mock.Mock(
            groupid="some_group",
            topic="topic1",
            partitions=None
        )
        expected_topics_dict = {"topic1": [0, 1, 2]}
        with self.mock_get_topics() as mock_get_topics:
            topics_dict = OffsetManagerBase.preprocess_args(
                args.groupid,
                args.topic,
                args.partitions,
                mock.Mock(),
                mock_kafka_client
            )
            assert mock_get_topics.called
            assert topics_dict == expected_topics_dict

    def test_preprocess_args_no_topics_no_partitions(self, mock_kafka_client):
        args = mock.Mock(
            groupid="some_group",
            topic=None,
            partitions=None
        )
        expected_topics_dict = {
            "topic1": [0, 1, 2],
            "topic2": [0, 1, 2, 3],
            "topic3": [0, 1],
        }
        with self.mock_get_topics() as mock_get_topics:
            topics_dict = OffsetManagerBase.preprocess_args(
                args.groupid,
                args.topic,
                args.partitions,
                mock.Mock(),
                mock_kafka_client
            )
            assert mock_get_topics.called
            assert topics_dict == expected_topics_dict

    def test_preprocess_args_wrong_topic(self, mock_kafka_client):
        args = mock.Mock(
            groupid="some_group",
            topic="topic34",
            partitions=None
        )
        with self.mock_get_topics() as mock_get_topics, mock.patch.object(
            sys,
            "exit",
            autospec=True,
        ) as mock_exit:
            OffsetManagerBase.preprocess_args(
                args.groupid,
                args.topic,
                args.partitions,
                mock.Mock(),
                mock_kafka_client
            )
            assert mock_get_topics.called
            assert mock_exit.called

    def test_preprocess_args_partitions_without_topic(self, mock_kafka_client):
        args = mock.Mock(
            groupid="some_group",
            topic=None,
            partitions=[0, 1, 2]
        )

        with self.mock_get_topics(), mock.patch.object(
            sys,
            "exit",
            autospec=True,
        ) as mock_exit:
            OffsetManagerBase.preprocess_args(
                args.groupid,
                args.topic,
                args.partitions,
                mock.Mock(),
                mock_kafka_client
            )
            assert mock_exit.called



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest
from mock import Mock
from mock import sentinel

from .helper import create_and_attach_partition
from .helper import create_broker
from kafka_utils.kafka_cluster_manager.cluster_info.broker import Broker
from kafka_utils.kafka_cluster_manager.cluster_info.error import \
    EmptyReplicationGroupError
from kafka_utils.kafka_cluster_manager.cluster_info.error import \
    NotEligibleGroupError
from kafka_utils.kafka_cluster_manager.cluster_info.rg import \
    ReplicationGroup
from kafka_utils.kafka_cluster_manager.cluster_info.topic import Topic


@pytest.fixture
def rg_unbalanced(create_partition):
    # Broker Topics:Partition
    #   1    topic1:0, topic1:1, topic4:0
    #   2    topic1:2, topic1:3, topic2:0, topic2:1, topic3:0
    #   3    topic4:0
    # total partitions: 9
    p10 = create_partition('topic1', 0)
    p11 = create_partition('topic1', 1)
    p12 = create_partition('topic1', 2)
    p13 = create_partition('topic1', 3)
    p20 = create_partition('topic2', 0)
    p21 = create_partition('topic2', 1)
    p30 = create_partition('topic3', 0)
    p40 = create_partition('topic4', 0, replication_factor=2)
    p50 = create_partition('topic5', 0)
    b1 = create_broker('b1', [p10, p12, p13, p21])
    b2 = create_broker('b2', [p11, p20, p30, p40])
    b3 = create_broker('b3', [p40])
    b4 = create_broker('b4', [p50])
    return ReplicationGroup('test_rg', set([b1, b2, b3, b4]))


@pytest.fixture
def rg_balanced(create_partition):
    # Broker Topics:Partition
    #   1    topic1:0, topic1:1, topic2:0, topic4:0
    #   2    topic1:2, topic1:3, topic2:1
    #   3    topic2:1, topic3:0, topic4:0
    # total partitions: 9
    p10 = create_partition('topic1', 0)
    p11 = create_partition('topic1', 1)
    p12 = create_partition('topic1', 2)
    p13 = create_partition('topic1', 3)
    p20 = create_partition('topic2', 0)
    p21 = create_partition('topic2', 1)
    p30 = create_partition('topic3', 0)
    p40 = create_partition('topic4', 0, replication_factor=2)
    b1 = create_broker('b1', [p10, p11, p20, p40])
    b2 = create_broker('b2', [p12, p13, p21])
    b3 = create_broker('b3', [p21, p30, p40])
    return ReplicationGroup('test_rg', set([b1, b2, b3]))


def assert_rg_balanced(rg):
    expected_count = len(rg.partitions) // len(rg.brokers)
    for broker in rg.brokers:
        if not broker.decommissioned:
            assert len(broker.partitions) in (expected_count, expected_count + 1)


class TestReplicationGroup(object):

    def test_rebalance_brokers(self, rg_unbalanced):
        orig_partitions = rg_unbalanced.partitions

        rg_unbalanced.rebalance_brokers()

        # No partitions are missing
        assert sorted(orig_partitions) == sorted(rg_unbalanced.partitions)
        assert_rg_balanced(rg_unbalanced)

    def test_rebalance_brokers_one_inactive(
        self,
        rg_unbalanced,
        create_partition,
    ):
        expected_count = \
            len(rg_unbalanced.partitions) // len(rg_unbalanced.brokers)
        p51 = create_partition('topic5', 1)
        b4 = create_broker('b4', [p51])
        b4.mark_inactive()
        rg_unbalanced.add_broker(b4)
        orig_partitions = rg_unbalanced.partitions

        rg_unbalanced.rebalance_brokers()

        # No partitions are missing
        assert sorted(orig_partitions) == sorted(rg_unbalanced.partitions)
        # b4 has not changed
        assert b4.partitions == set([p51])
        for broker in rg_unbalanced.brokers:
            if not broker.decommissioned and not broker.inactive:
                assert len(broker.partitions) in (
                    expected_count,
                    expected_count + 1
                )

    def test_rebalance_brokers_balanced(self, rg_balanced):
        expected = {b: b.partitions for b in rg_balanced.brokers}

        rg_balanced.rebalance_brokers()

        # there shouldn't be any partition movement
        assert expected == {b: b.partitions for b in rg_balanced.brokers}
        assert_rg_balanced(rg_balanced)

    def test_rebalance_decommissioned_broker(
        self,
        rg_balanced,
        create_partition,
    ):
        # rg_balanced is supposed to have 3 brokers with 3 partitions each. We
        # add another broker with 3 more partitions
        broker_count = len(rg_balanced.brokers)
        p50 = create_partition('topic5', 0)
        p51 = create_partition('topic5', 1)
        p52 = create_partition('topic5', 2)
        broker = create_broker('broker4', [p50, p51, p52])
        rg_balanced.add_broker(broker)

        broker.mark_decommissioned()
        orig_partitions = rg_balanced.partitions

        rg_balanced.rebalance_brokers()

        assert broker.empty() is True
        assert sorted(orig_partitions) == sorted(rg_balanced.partitions)
        expected_count = len(rg_balanced.partitions) // broker_count
        for broker in rg_balanced.brokers:
            if not broker.decommissioned:
                assert len(broker.partitions) in (
                    expected_count,
                    expected_count + 1,
                )

    def test_rebalance_new_empty_broker(self, rg_balanced):
        broker = Broker('4')
        rg_balanced.add_broker(broker)
        orig_partitions = rg_balanced.partitions

        rg_balanced.rebalance_brokers()

        assert sorted(orig_partitions) == sorted(rg_balanced.partitions)
        assert_rg_balanced(rg_balanced)

    def test_rebalance_balanced_inactive_broker(self, rg_balanced):
        expected = {b: b.partitions for b in rg_balanced.brokers}
        list(rg_balanced.brokers)[0].mark_inactive()

        rg_balanced.rebalance_brokers()

        assert expected == {b: b.partitions for b in rg_balanced.brokers}
        assert_rg_balanced(rg_balanced)

    def test_decommission_no_remaining_brokers(self, create_partition):
        p10 = create_partition('topic1', 0)
        p11 = create_partition('topic1', 1)
        p20 = create_partition('topic2', 0)
        b1 = create_broker('b1', [p10, p11, p20])
        b2 = create_broker('b2', [])
        b2.mark_inactive()
        rg = ReplicationGroup('rg', set([b1, b2]))
        b1.mark_decommissioned()

        # Two brokers b1 decommissioned b2 inactive
        with pytest.raises(EmptyReplicationGroupError):
            rg.rebalance_brokers()

    def test_decommission_not_enough_replicas(self, create_partition):
        p10 = create_partition('topic1', 0)
        p11 = create_partition('topic1', 1)
        p20 = create_partition('topic2', 0, replication_factor=2)
        b1 = create_broker('b1', [p10, p11, p20])
        b2 = create_broker('b1', [p20])
        rg = ReplicationGroup('rg', set([b1, b2]))
        b2.mark_decommissioned()

        rg.rebalance_brokers()

        assert not b2.empty()

    def test_rebalance_empty_replication_group(self):
        rg = ReplicationGroup('empty_rg')

        with pytest.raises(EmptyReplicationGroupError):
            rg.rebalance_brokers()

    def test_decommission_an_inactive_broker(
        self,
        rg_balanced,
        create_partition,
    ):
        # rg_balanced is supposed to have 3 brokers with 3 partitions each. We
        # add another broker with 3 more partitions
        broker_count = len(rg_balanced.brokers)
        p50 = create_partition('topic5', 0)
        p51 = create_partition('topic5', 1)
        p52 = create_partition('topic5', 2)
        broker = create_broker('broker4', [p50, p51, p52])
        rg_balanced.add_broker(broker)

        broker.mark_decommissioned()
        broker.mark_inactive()
        orig_partitions = rg_balanced.partitions

        rg_balanced.rebalance_brokers()

        assert broker.empty() is True
        assert sorted(orig_partitions) == sorted(rg_balanced.partitions)
        expected_count = len(rg_balanced.partitions) // broker_count
        for broker in rg_balanced.brokers:
            if not broker.decommissioned:
                assert len(broker.partitions) in (
                    expected_count,
                    expected_count + 1,
                )

    def test_add_broker_empty(self):
        rg = ReplicationGroup('test_rg', None)
        rg.add_broker(sentinel.broker)

        assert set([sentinel.broker]) == rg.brokers

    def test_invalid_brokers_type_list(self):
        with pytest.raises(TypeError):
            ReplicationGroup('test_rg', [sentinel.broker1, sentinel.broker2])

    def test_add_broker(self):
        rg = ReplicationGroup(
            'test_rg',
            set([sentinel.broker1, sentinel.broker2]),
        )
        rg.add_broker(sentinel.broker)

        assert sentinel.broker in rg.brokers

    def test_id(self):
        rg = ReplicationGroup('test_rg', None)

        assert 'test_rg' == rg.id

    def test_partitions(self):
        mock_brokers = set([
            Mock(spec=Broker, partitions=set([sentinel.p1, sentinel.p2])),
            Mock(spec=Broker, partitions=set([sentinel.p3, sentinel.p1])),
        ])
        rg = ReplicationGroup('test_rg', mock_brokers)
        expected = [
            sentinel.p1,
            sentinel.p2,
            sentinel.p3,
            sentinel.p1,
        ]

        assert sorted(expected) == sorted(rg.partitions)

    def test_acquire_partition(self, create_partition):
        p10 = create_partition('t1', 0)
        p11 = create_partition('t1', 1)
        p12 = create_partition('t1', 2)
        p13 = create_partition('t1', 3)
        p20 = create_partition('t2', 0)
        p21 = create_partition('t2', 1)
        b1 = create_broker('b1', [p10, p11, p20])
        b2 = create_broker('b2', [p12, p21])
        rg = ReplicationGroup('test_rg', set([b1, b2]))
        b3 = create_broker('b3', [p13])

        rg.acquire_partition(p13, b3)

        assert b3.empty()
        assert p13 in b2.partitions

    def test_acquire_partition_error(self, create_partition):
        p10 = create_partition('t1', 0)
        p11 = create_partition('t1', 1)
        p20 = create_partition('t2', 0)
        b1 = create_broker('b1', [p10, p11, p20])
        rg = ReplicationGroup('test_rg', set([b1]))
        b3 = create_broker('b3', [p11])

        # Not eligible because broker b3 has already a replica of partition p11
        with pytest.raises(NotEligibleGroupError):
            rg.acquire_partition(p11, b3)

    def test__elect_source_broker(self, create_partition):
        p10 = create_partition('t1', 0)
        p20 = create_partition('t2', 0)
        p11 = create_partition('t1', 1)
        b1 = create_broker('b1', [p10, p11])  # b1 -> t1: {0,1}, t2: 0
        p30 = create_partition('t3', 0)
        b2 = create_broker('b2', [p10, p20, p30])  # b2 -> t1: 0, t2: 1, t3: 0
        rg = ReplicationGroup('test_rg', set([b1, b2]))

        # b1 has 2 partitions (p1 and p3) for same topic t1
        # b2 has only 1 partition (p1) for topic t1
        # source broker should be b1 to reduce the number of partitions of the
        # same topic
        victim_partition = p10
        actual = rg._elect_source_broker(victim_partition)
        assert actual == b1

    def test__elect_dest_broker(self, create_partition):
        p10 = create_partition('t1', 0)
        p11 = create_partition('t1', 1)
        p20 = create_partition('t2', 0)
        p30 = create_partition('t3', 0)
        b1 = create_broker('b1', [p10, p20, p11])  # b1 -> t1: {0,1}, t2: 0
        b2 = create_broker('b2', [p10, p30])  # b2 -> t1: 0, t3: 0
        rg = ReplicationGroup('test_rg', set([b1, b2]))

        # Since p30 is already in b2 so the preferred destination will be b1
        # although b2 has less partitions.
        victim_partition = p30
        actual = rg._elect_dest_broker(victim_partition)
        assert actual == b1

    def test__elect_dest_broker_prefer_less_siblings(self, create_partition):
        p10 = create_partition('t1', 0)
        p11 = create_partition('t1', 1)
        p20 = create_partition('t2', 0)
        b1 = create_broker('b1', [p10, p11, p20])  # b1 -> t1: {0,1}, t2: 0
        p30 = create_partition('t3', 0)
        p31 = create_partition('t3', 1)
        p32 = create_partition('t3', 2)
        b2 = create_broker('b2', [p10, p30, p31, p32])  # b2 -> t1: 0, t3: 0
        rg = ReplicationGroup('test_rg', set([b1, b2]))

        # Since t1 has two partitions in b1 but only one in b2,
        # the preferred destination should be b2
        victim_partition_p12 = create_partition('t1', 2)
        actual = rg._elect_dest_broker(victim_partition_p12)
        assert actual == b2

    def test__elect_dest_broker_partition_conflict(self, create_partition):
        p1 = create_partition('t1', 0)
        p2 = create_partition('t2', 0)
        p3 = create_partition('t1', 1)
        b1 = create_broker('b1', [p1, p2, p3])  # b1 -> t1: {0,1}, t2: 0
        rg = ReplicationGroup('test_rg', set([b1]))
        # p1 already exists in b1
        # This should never happen and we expect the application to fail badly
        victim_partition = p1
        assert None is rg._elect_dest_broker(victim_partition)

    def test_count_replica(self, create_partition):
        p10 = create_partition('t1', 0)
        p11 = create_partition('t1', 1)
        p12 = create_partition('t1', 2)
        p13 = create_partition('t1', 3)
        b1 = create_broker('b1', [p10, p11, p12])
        b2 = create_broker('b2', [p10, p13])
        rg = ReplicationGroup('test_rg', set([b1, b2]))

        assert rg.count_replica(p10) == 2
        assert rg.count_replica(p11) == 1
        assert rg.count_replica(p12) == 1
        assert rg.count_replica(p13) == 1
        assert rg.count_replica(create_partition('t1', 4)) == 0

    def test__select_broker_pair(self, create_partition):
        p10 = create_partition('t1', 0)
        p11 = create_partition('t1', 1)
        p20 = create_partition('t2', 0)
        b0 = create_broker('b0', [p10, p20, p11])
        b1 = create_broker('b1', [p10, p20])
        b2 = create_broker('b2', [p20, p11])
        b3 = create_broker('b3', [p10, p20, p11])
        rg_source = ReplicationGroup('rg1', set([b0, b1, b2, b3]))
        b4 = create_broker('b4', [p20, p11])
        b5 = create_broker('b5', [p20])
        b6 = create_broker('b6', [p10])
        b7 = create_broker('b7', [p11])
        rg_dest = ReplicationGroup('rg2', set([b4, b5, b6, b7]))

        # Select best-suitable brokers for moving partition 'p10'
        broker_source, broker_dest = rg_source._select_broker_pair(rg_dest, p10)

        # source-broker can't be b2 since it doesn't have p10
        # source-broker shouldn't be b1 since it has lesser partitions
        # than b3 and b0
        assert broker_source in [b0, b3]
        # dest-broker shouldn't be b4 since it has more partitions than b4
        # dest-broker can't be b6 since it has already partition p10
        assert broker_dest in [b5, b7]

    def test_move_partition(self, create_partition):
        p10 = create_partition('t1', 0)
        p11 = create_partition('t1', 1)
        p20 = create_partition('t2', 0)
        b1 = create_broker('b1', [p10, p20, p11])
        b2 = create_broker('b2', [p10, p11])
        # 2 p1 replicas are in rg_source
        rg_source = ReplicationGroup('rg1', set([b1, b2]))
        b3 = create_broker('b3', [p10, p11])
        b4 = create_broker('b4', [p20])
        # 1 p1 replica is in rg_dest
        rg_dest = ReplicationGroup('rg2', set([b3, b4]))

        # Move partition p1 from rg1 to rg2
        rg_source.move_partition(rg_dest, p10)

        # partition-count of p1 for rg1 should be 1
        assert rg_source.count_replica(p10) == 1
        # partition-count of p1 for rg2 should be 2
        assert rg_dest.count_replica(p10) == 2
        # Rest of the partitions are untouched
        assert rg_source.count_replica(p20) == 1
        assert rg_dest.count_replica(p20) == 1
        assert rg_source.count_replica(p11) == 2
        assert rg_dest.count_replica(p11) == 1

    def test__get_target_brokers_1(self, create_partition):
        p10 = create_partition('topic1', 0, replication_factor=2)
        p11 = create_partition('topic1', 1, replication_factor=2)
        p20 = create_partition('topic2', 0, replication_factor=2)
        p30 = create_partition('topic3', 0)
        b1 = create_broker('b1', [p10, p11, p20, p30])
        b2 = create_broker('b2', [p10, p20])
        rg = ReplicationGroup('rg', set([b1, b2]))

        over_loaded = [b1]
        under_loaded = [b2]
        target = rg._get_target_brokers(
            over_loaded,
            under_loaded,
            rg.generate_sibling_distance(),
        )

        # p30 is the best fit because topic3 doesn't have any partition in
        # broker 2
        assert target == (b1, b2, p30) or target == (b1, b2, p11)

    def test_get_target_brokers_2(self, create_partition):
        p10 = create_partition('topic1', 0, replication_factor=2)
        p11 = create_partition('topic1', 1, replication_factor=2)
        p20 = create_partition('topic2', 0, replication_factor=1)
        p21 = create_partition('topic2', 1, replication_factor=1)
        p30 = create_partition('topic3', 0)
        b1 = create_broker('b1', [p10, p11, p20])
        b2 = create_broker('b2', [p10, p30])
        b3 = create_broker('b3', [p21])
        rg = ReplicationGroup('rg', set([b1, b2, b3]))

        over_loaded = [b1]
        under_loaded = [b2, b3]

        target = rg._get_target_brokers(
            over_loaded,
            under_loaded,
            rg.generate_sibling_distance(),
        )

        # b3 is the best broker fit because of number of partition
        # Both p10 and p11 are a good fit. p20 can't be considered because it is
        # already in b3.
        assert target == (b1, b3, p10) or target == (b1, b3, p11)

    def test_get_target_brokers_3(self, create_partition):
        p10 = create_partition('topic1', 0)
        p11 = create_partition('topic1', 1)
        p12 = create_partition('topic1', 2)
        p20 = create_partition('topic2', 0)
        p21 = create_partition('topic2', 1)
        p30 = create_partition('topic3', 0)
        p31 = create_partition('topic3', 1)
        b1 = create_broker('b1', [p10, p11, p20])
        b2 = create_broker('b2', [p12, p21, p30, p31])
        b3 = create_broker('b3', [])
        rg = ReplicationGroup('rg', set([b1, b2, b3]))

        over_loaded = [b1, b2]
        under_loaded = [b3]

        target = rg._get_target_brokers(
            over_loaded,
            under_loaded,
            rg.generate_sibling_distance(),
        )

        # b3 is the best broker fit because of number of partition
        # Both p10 and p11 are a good fit. p20 can't be considered because it is
        # already in b3.
        assert target == (b2, b3, p30) or target == (b2, b3, p31)

    def test_generate_sibling_count(self):
        t1 = Topic('topic1', 2)
        t2 = Topic('topic2', 2)
        t3 = Topic('topic3', 1)
        p10 = create_and_attach_partition(t1, 0)
        p11 = create_and_attach_partition(t1, 1)
        p12 = create_and_attach_partition(t1, 2)
        p20 = create_and_attach_partition(t2, 0)
        p21 = create_and_attach_partition(t2, 1)
        p22 = create_and_attach_partition(t2, 2)
        p30 = create_and_attach_partition(t3, 0)
        p31 = create_and_attach_partition(t3, 1)
        b1 = create_broker('b1', [p10, p11, p20, p21, p30, p31])
        b2 = create_broker('b2', [p12, p21, p22])
        b3 = create_broker('b3', [p10, p11, p22])
        rg = ReplicationGroup('rg', set([b1, b2, b3]))

        expected = {
            b1: {b2: {t1: 1, t2: 0}, b3: {t1: 0, t2: 1}},
            b2: {b1: {t1: -1, t2: 0, t3: -2}, b3: {t1: -1, t2: 1}},
            b3: {b1: {t1: 0, t2: -1, t3: -2}, b2: {t1: 1, t2: -1}},
        }
        actual = rg.generate_sibling_distance()

        assert dict(actual) == expected

    def test_update_sibling_count(self):
        t1 = Topic('topic1', 2)
        t2 = Topic('topic2', 2)
        t3 = Topic('topic3', 1)
        p10 = create_and_attach_partition(t1, 0)
        p11 = create_and_attach_partition(t1, 1)
        p12 = create_and_attach_partition(t1, 2)
        p20 = create_and_attach_partition(t2, 0)
        p21 = create_and_attach_partition(t2, 1)
        p22 = create_and_attach_partition(t2, 2)
        p30 = create_and_attach_partition(t3, 0)
        p31 = create_and_attach_partition(t3, 1)
        b1 = create_broker('b1', [p10, p11, p20, p21, p30, p31])
        b2 = create_broker('b2', [p12, p21, p22])
        b3 = create_broker('b3', [p10, p11, p22])
        rg = ReplicationGroup('rg', set([b1, b2, b3]))
        sibling_distance = {
            b2: {
                b1: {t1: -1, t2: 0, t3: -2},
                b3: {t1: 1, t2: -1, t3: -2}
            },
        }
        # Move a p10 from b1 to b2
        b1.move_partition(p10, b2)

        # NOTE: b2: b1: t1: -1 -> 1 and b2: b3: t1: 1 -> 0
        expected = {
            b2: {
                b1: {t1: 1, t2: 0, t3: -2},
                b3: {t1: 0, t2: -1, t3: -2},
            },
        }
        actual = rg.update_sibling_distance(sibling_distance, b2, t1)

        assert dict(actual) == expected



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import print_function

import json
import sys
from itertools import groupby


def groupsortby(data, key):
    """Sort and group by the same key."""
    return groupby(sorted(data, key=key), key)


def dict_merge(set1, set2):
    """Joins two dictionaries."""
    return dict(set1.items() + set2.items())


def to_h(num, suffix='B'):
    """Converts a byte value in human readable form."""
    if num is None:  # Show None when data is missing
        return "None"
    for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)


def to_int(num):
    """
    Converts 'num' to int representation in string
    or to "None" in case of None.
    """
    if num is None:
        return "None"
    return "{:.0f}".format(num)


def to_float(num):
    """
    Converts 'num' to float representation in string
    or to "None" in case of None.
    """
    if num is None:
        return "None"
    return "{:.2f}".format(num)


def format_to_json(data):
    """Converts `data` into json
    If stdout is a tty it performs a pretty print.
    """
    if sys.stdout.isatty():
        return json.dumps(data, indent=4, separators=(',', ': '))
    else:
        return json.dumps(data)


def print_json(data):
    """Converts `data` into json and prints it to stdout."""
    print(format_to_json(data))
